with Flts, Globals, Params, Strings, Reps;
use Flts, Globals, Params, Strings, Reps;
with Params.Names, Matrices, Vectors, ScalVectors, Legends, Legends.Chain, Legends.Chain.Fix;

pragma Elaborate_All (Flts,Globals,Params,Strings,Reps);
pragma Elaborate_All (Params.Names,Matrices,Vectors,ScalVectors,Legends.Chain.Fix);

procedure Approx_Fixpt(P: in Param) is

  package PN is new Params.Names(P => P, Trunc => True);
  use PN;

  subtype Scalar is Rep;

  NumDigits: constant Positive := Reps.Set_Precision(256);

  package SV is new Vectors (Component => Scalar);
  package SLS is new Legends (CLast => CLast, Rho => Rho, Scalar => Scalar, SV => SV);
  package SLSC is new SLS.Chain(JCMax => JCMax, Scale => Scale);
  package SM is new Matrices (Scalar => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package Fix is new SLSC.Fix(AAPow => AAPow, Mat_PackSize => Mat_PackSize, Decay => Decay, Mu => Mu, Model => Model, Symm => Symm, SM => SM, SSV => SSV);

  use SLS,SLSC,SM,Fix;

  JCut: constant Positive := Integer'Max(JCMax/2,8);
  Eps: constant Radius := 1.0e-15;

  F: LChain;
  M: QC_Matrix;

begin

  Verbosity := 0;
  Show("",P);

  Read(Input_Dir & Fix_Name,F);
  F.W := Decay;
  ApproxFix(F,M,Eps);
  --IterContr(F,M);
  Show1("norm F",MaxNorm(F));
  Write(Output_Dir & Mat_Name,M);
  Write(Output_Dir & Fix_Name,F);
  WritePlot(Output_Dir & Plot_Name,F,JCut);
  WritePlotSum(Output_Dir & PlotSum_Name,F,JCut);

end Approx_Fixpt;
