with Primes;
use Primes;

pragma Elaborate_All (Primes);

package body Binoms is

  procedure Num_Mult(B: in BigRat; S,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    SPrime: Int_Scalar renames Tmp3;
    SPower: Int_Scalar renames Tmp4;
    BNum: Integer := Integer'Max(0,B(1));
    R: Flt;
  begin
    Mult(Two**BNum,S,Tmp5);
    for I in 2 .. Size loop
      BNum := Integer'Max(0,B(I));
      if BNum>1 then
        R := Flt(Prime(I))**BNum;
        if R<IntegerLast then -- result is exact
          Mult(R,S,Tmp5);
        else
          Assign(Prime(I),SPrime);
          IPower(BNum,SPrime,SPower,Tmp5);
          Mult(SPower,S);
        end if;
      elsif BNum=1 then
        Mult(Flt(Prime(I)),S,Tmp5);
      end if;
    end loop;
  end Num_Mult;

  procedure Den_Mult(B: in BigRat; S,Tmp3,Tmp4,Tmp5: in out Int_Scalar) is
    SPrime: Int_Scalar renames Tmp3;
    SPower: Int_Scalar renames Tmp4;
    BDen: Integer := Integer'Max(0,-B(1));
    R: Flt;
  begin
    Mult(Two**BDen,S,Tmp5);
    for I in 2 .. Size loop
      BDen := Integer'Max(0,-B(I));
      if BDen>1 then
        R := Flt(Prime(I))**BDen;
        if R<IntegerLast then -- result is exact
          Mult(R,S,Tmp5);
        else
          Assign(Prime(I),SPrime);
          IPower(BDen,SPrime,SPower,Tmp5);
          Mult(SPower,S);
        end if;
      elsif BDen=1 then
        Mult(Flt(Prime(I)),S,Tmp5);
      end if;
    end loop;
  end Den_Mult;

  function Make_Binom return Binom_Type is
    --- Binom(N)=(2N choose N)
    Factorial: constant Factorial_Type := Make_Factorials;
    B: BigRat;
    SNum,SDen,Tmp3,Tmp4,Tmp5: Int_Scalar;
    Binom: Binom_Type;
  begin
    Assign(1,Binom(0));
    Assign(2,Binom(1));
    for N in 2 .. Binom'Last loop
      Quot(Factorial(2*N),Factorial(N),B);
      Div(Factorial(N),B);
      Assign(1,SNum);
      Num_Mult(B,SNum,Tmp3,Tmp4,Tmp5);
      Assign(1,SDen);
      Den_Mult(B,SDen,Tmp3,Tmp4,Tmp5);
      Quot(SNum,SDen,Binom(N));
    end loop;
    return Binom;
  end Make_Binom;

end Binoms;
