with Strings, Globals;
use Strings, Globals;

pragma Elaborate_All (Strings,Globals);

package body Legends.Chain.Fix is

  use Lin,Lin_Contr;

  procedure SetZero(Par: in Parity; Q: in out LChain) is
  begin
    SetZero(Decay,Par,Q);
  end SetZero;

  procedure SetZero(Q: in out LChain) is
  begin
    SetZero(Decay,Symm,Q);
  end SetZero;

  procedure Power(F: in LChain; PF,Tmp: in out LChain; M: in Positive := Model) is
    --- PF := F**M1
    M1: constant Natural := M-1;
  begin
    if M1=1 then
      Copy(F,PF);
    elsif M1=2 then
      Prod(F,F,PF);
    elsif M1=3 then
      Prod(F,F,Tmp);
      Prod(Tmp,F,PF);
    elsif M1=4 then
      Prod(F,F,Tmp);
      Prod(Tmp,Tmp,PF);
    else
      raise Not_Implemented;
    end if;
  end Power;

  procedure GMap(F: in LChain; GF: in out LChain) is
    --- GF := Nu*SumAAPowers(K,F**Model)+(Mu**K)*AAPower(K,F), K=AAPow
    Tmp: LChain;
  begin
    Power(F,GF,Tmp);
    SumAAPowers(AAPow,GF,Decay,SMu);
    Copy(F,Tmp);
    AAPower(AAPow,Tmp,Decay);
    AddProd(MuAAPow,Tmp,GF);
  end GMap;

  function DPower(F: LChain; M: in Positive := Model) return LChain is
    --- DPF := M1*F**(M1-1);
    M1: constant Natural := M-1;
    S: constant Scalar := Scal(M1);
    DPF: LChain;
  begin
    if M1=1 then
      SetZero(0,DPF);
      Copy(S,DPF.A);
    elsif M1=2 then
      Prod(S,F,DPF);
    elsif M1=3 then
      Prod(F,F,DPF);
      Mult(S,DPF);
    else
      declare
        Tmp: LChain;
      begin
        Power(F,DPF,Tmp,M-1);
        Mult(S,DPF);
      end;
    end if;
    return DPF;
  end DPower;

  procedure DGMap(DPF,H: in LChain; DGH: in out LChain) is
    --- DGH := Nu*SumAAPowers(K,DFP*H)+(Mu**K)*AAPower(K,H), K=AAPow
    Tmp: LChain;
  begin
    Prod(DPF,H,DGH);
    SumAAPowers(AAPow,DGH,Decay,SMu);
    Copy(H,Tmp);
    AAPower(AAPow,Tmp,Decay);
    AddProd(MuAAPow,Tmp,DGH);
  end DGMap;

  procedure DContrNorm(M: in SMatrix; F: in LChain; PackSize: in Positive; R: out Radius) is
    Q: constant LModes := Make(PackSize,Decay,Symm,True,False); --- Err modes, but Extract is not used
    DPF: constant LChain := DPower(F);
    procedure DG(H: in LChain; DGH: in out LChain) is begin DGMap(DPF,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in LChain; DCH: in out LChain) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if PackSize <= Mat_PackSize then raise Not_Implemented; end if;
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,R);
  end DContrNorm;

  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out LChain) is
    KMax: constant Radius := Flt(7)/Flt(8);
    E,K: Radius;
    ANorm,R,Eps,Del: Scalar;
    F0: LChain;
  begin
    Show0("ContrFix: ",False);
    NumTrunc(F,F0);
    GMap(F0,F);
    Sub(F0,F);
    E := MaxNorm(F);
    Assign(E,Eps);
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",E,Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(F0,F);
    AddBall(R,F);
    DContrNorm(M,F,PackSize,K);
    Show1("ContrFix: K =",K);
    if K>KMax then raise Sorry with "K too large"; end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(F0,F);
    AddBall(R,F);
  end ContrFix;

  procedure Show_Extrema(W: in FltVec) is
    Fac: constant Radius := Two**12;
    L: constant Integer := W'Length;
    WK,WLast,WMin,WMax: Flt := Zero;
  begin
    Show1("extrema count:",L);
    Show0("bumps: ",False);
    for K in W'Range loop
      WK := W(K);
      if WK>WLast then
        if WK>Zero then Show0("^",False); end if;
        WMax := RMax(WK,WMax);
      elsif WK<WLast then
        if WK<Zero then Show0("v",False); end if;
        WMin := RMin(WK,WMin);
      else
        raise Not_Implemented;
      end if;
      WLast := WK;
    end loop;
    Show0;
    Show1("Min <=",Flt'Ceiling(WMin*Fac)/Fac);
    Show1("Max >=",Flt'Floor(WMax*Fac)/Fac);
  end Show_Extrema;

  procedure Show_TeX(W: in FltVec) is
    WK,WLast: Flt := Zero;
  begin
    Show0("TeX: ",False);
    for K in W'Range loop
      WK := W(K);
      if WK>Zero then
        if WK>WLast then
          Show0("\PPP",False);
        elsif WK<WLast then
          Show0("\ppp",False);
        else
          raise Not_Implemented;
        end if;
      elsif WK<Zero then
        if WK<WLast then
          Show0("\NNN",False);
        elsif WK>WLast then
          Show0("\nnn",False);
        else
          raise Not_Implemented;
        end if;
      else
        raise Not_Implemented;
      end if;
      WLast := WK;
    end loop;
    Show0;
  end Show_TeX;

  procedure Show_Stat(Q: in LChain) is
    BCut: constant Radius := Half**6;
    W: constant FltVec := UpDown(Q,BCut);
  begin
    Show_Extrema(W);
    Show_Tex(W);
  end Show_Stat;

  --------------------------------------------
  --------------------------------------------

  procedure Unpack(V: in SVector; F: in out LChain) is
  begin
    SetZero(F);
    AddProd(QC,V,F);
  end Unpack;

  procedure DerMatrix(F: in LChain; M: in out SMatrix) is
    --- for numeric mode only
    DPF: constant LChain := DPower(F);
    procedure DG(H: in LChain; DGH: in out LChain) is begin DGMap(DPF,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_STrunc then raise Sorry; end if;
    Enclose(QC,M);
  end DerMatrix;

  procedure ContrMatrix(F: in LChain; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(F,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ApproxFix(F: in out LChain; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Contract0 is new Contr0 (Map => GMap);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 8;  --- was 4
    Fac: constant Radius := One;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    G0,G: LChain;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(F,G);
    for I in 1 .. Iter loop
      ContrMatrix(G,M);
      Copy(G,G0);
      Contract0(QC,M,G,Err);
      NumInterpolate(G0,G,Fac);  --- G := (1-Fac)*G0+Fac*G
      Show1("ApproxFix: step " & Strng(I) & ", Err ",Err);
      if Err<MinErr then
        Copy(G0,F);              --- was Copy(G,F), and G0 did not exist
        MinErr := Err;
        if Err<Eps then exit; end if;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: MinErr ",MinErr);
    ContrMatrix(F,M);
  end ApproxFix;

  procedure IterContr(F: in out LChain; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new Contr0 (Map => GMap);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G: LChain;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    ContrMatrix(G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
        ContrMatrix(G,M);
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
    ContrMatrix(F,M);
  end IterContr;

end Legends.Chain.Fix;
