with ScalVectors, Matrices, Linear, Linear.Contr;

pragma Elaborate_All (Matrices,ScalVectors,Linear.Contr);

generic

  AAPow: in Positive;
  Mat_PackSize: in Positive;
  Decay: in Radius;
  Mu: in Flt;
  Symm: in Parity;
  Model: in Positive;

  with package SM is new Matrices(Scalar => Scalar);
  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package Legends.Chain.Fix is

  subtype SMatrix is SM.Matrix;

  --- for traveling-waves
  procedure SetZero(Par: in Parity; Q: in out LChain); --- Q.W=Decay
  procedure SetZero(Q: in out LChain);                 --- Q.W=Decay, Q.Par=Symm
  procedure Power(F: in LChain; PF,Tmp: in out LChain; M: in Positive := Model);
  procedure GMap(F: in LChain; GF: in out LChain);
  function DPower(F: LChain; M: in Positive := Model) return LChain;
  procedure DGMap(DPF,H: in LChain; DGH: in out LChain);
  procedure DContrNorm(M: in SMatrix; F: in LChain; PackSize: in Positive; R: out Radius);
  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out LChain);

  --- info
  procedure Show_Extrema(W: in FltVec);
  procedure Show_TeX(W: in FltVec);
  procedure Show_Stat(Q: in LChain);

  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => LChain, Mode => LMode, Modes => LModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  --- numerical
  procedure Unpack(V: in SVector; F: in out LChain);
  procedure DerMatrix(F: in LChain; M: in out SMatrix);
  procedure ContrMatrix(F: in LChain; M: in out SMatrix);
  procedure ApproxFix(F: in out LChain; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(F: in out LChain; M: in out SMatrix);

  QC: constant LModes := Make(Mat_PackSize,Decay,Symm,True,True);   --- Simple and no Err modes
  subtype QC_Vector is SVector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

  DContr_Parallel: Boolean renames Lin.Op_Norm_Parallel;
  DerMatrix_Parallel: Boolean renames Lin.Enclosure_Parallel;

  Tiny:    constant Radius := 1.0E-999;
  SMu:     constant Scalar := Scal(Mu)+BallAt0(Tiny);
  STwo:    constant Scalar := Scal(2);
  SThree:  constant Scalar := Scal(3);
  MuAAPow: constant Scalar := SMu**AAPow;

end Legends.Chain.Fix;
