with Strings, Protected_Counters, Globals, FltChains;
use Strings, Protected_Counters, Globals;

pragma Elaborate_All (Strings,Protected_Counters,Globals,FltChains);

package body Legends.Chain is

  use SV,LV;

  protected body Protected_Powers is
    function Old_Power(X: Radius; J: JERange) return Radius is
    begin
      if X=A(1) then return A(J); end if;
      if X=B(1) then return B(J); end if;
      return Zero;
    end Old_Power;
    procedure New_Power(X: in Radius; J: in JERange; Y: out Radius) is
    begin
      B := A;
      A(1) := X;
      for N in 2 .. JEMax loop A(N) := A(N/2)*A(N-N/2); end loop;
      Y := A(J);
    end New_Power;
  end Protected_Powers;

  function Prot_Power(X: Radius; J: JERange) return Radius is
    Y: Radius := Prot.Old_Power(X,J);
  begin
    if Y=Zero then Prot.New_Power(X,J,Y); end if;
    return Y;
  end Prot_Power;

  -------------------- chains

  function Consistent(Q: LChain) return Boolean is
    C: LGerms renames Q.C;
  begin
    if Q.W<One then return False; end if;
    if not (Q.Par=0 or else IsZero(Q.A)) then return False; end if;
    if not CheckPar(Q.Par,C(0)) then return False; end if;
    return STrunc or else Center0(C(JEMax));
  end Consistent;

  procedure SetZero(W: in Radius; Par: in Parity; Q: in out LChain) is
  begin
    Q.W := W;
    Q.Par := Par;
    SetZero(Q.A);
    SetZero(Q.C);
  end SetZero;

  procedure Copy(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Copy(Q1.A,Q2.A);
    for J in 0 .. JLast loop Copy(C1(J),C2(J)); end loop;
  end Copy;

  procedure Swap(Q1,Q2: in out LChain) is
    P1: constant Parity := Q1.Par;
    W1: constant Radius := Q1.W;
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q1.Par := Q2.Par; Q2.Par := P1;
    Q1.W := Q2.W; Q2.W := W1;
    Swap(Q1.A,Q2.A);
    for J in 0 .. JLast loop Swap(C1(J),C2(J)); end loop;
  end Swap;

  function IsZero(Q: LChain) return Boolean is
    C: LGerms renames Q.C;
  begin
    if not IsZero(Q.A) then return False; end if;
    for J in 0 .. JLast loop
      if not IsZero(C(J)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  function IsFinite(Q: LChain) return Boolean is
  begin
    if not IsZero(Q.A) then return False; end if;
    if STrunc then return True; end if;
    return IsZero(Q.C(JEMax));
  end IsFinite;

  procedure Neg(Q: in out LChain) is
    C: LGerms renames Q.C;
  begin
    Neg(Q.A);
    for J in JCRange loop Neg(C(J)); end loop;
  end Neg;

  procedure Neg(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Neg(Q1.A,Q2.A);
    for J in 0 .. JLast loop Neg(C1(J),C2(J)); end loop;
  end Neg;

  procedure Add(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    if Q2.W>Q1.W then Q2.W := Q1.W; end if;
    if Q1.Par /= Q2.Par then raise Parity_Error; end if;
    Add(Q1.A,Q2.A);
    for J in 0 .. JLast loop Add(C1(J),C2(J)); end loop;
  end Add;

  procedure Sum(Q1,Q2: in LChain; Q3: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
    C3: LGerms renames Q3.C;
  begin
    Q3.W := RMin(Q1.W,Q2.W);
    if Q1.Par /= Q2.Par then raise Parity_Error; end if;
    Q3.Par := Q1.Par;
    Sum(Q1.A,Q2.A,Q3.A);
    for J in 0 .. JLast loop Sum(C1(J),C2(J),C3(J)); end loop;
  end Sum;

  procedure Sub(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    if Q2.W>Q1.W then Q2.W := Q1.W; end if;
    if Q1.Par /= Q2.Par then raise Parity_Error; end if;
    Sub(Q1.A,Q2.A);
    for J in 0 .. JLast loop Sub(C1(J),C2(J)); end loop;
  end Sub;

  procedure Diff(Q1,Q2: in LChain; Q3: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
    C3: LGerms renames Q3.C;
  begin
    Q3.W := RMin(Q1.W,Q2.W);
    if Q1.Par /= Q2.Par then raise Parity_Error; end if;
    Q3.Par := Q1.Par;
    Diff(Q1.A,Q2.A,Q3.A);
    for J in 0 .. JLast loop Diff(C1(J),C2(J),C3(J)); end loop;
  end Diff;

  procedure Mult(R: in Flt; Q: in out LChain) is
    C: LGerms renames Q.C;
  begin
    Mult(R,Q.A);
    for J in 0 .. JLast loop Mult(R,C(J)); end loop;
  end Mult;

  procedure Prod(R: in Flt; Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Prod(R,Q1.A,Q2.A);
    for J in 0 .. JLast loop Prod(R,C1(J),C2(J)); end loop;
  end Prod;

  function "*"(R: Flt; Q: LChain) return LChain is
    Qt: LChain;
  begin
    Prod(R,Q,Qt);
    return Qt;
  end "*";

  procedure Mult(S: in Scalar; Q: in out LChain) is
    C: LGerms renames Q.C;
  begin
    Mult(S,Q.A);
    for J in 0 .. JLast loop Mult(S,C(J)); end loop;
  end Mult;

  procedure Prod(S: in Scalar; Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Prod(S,Q1.A,Q2.A);
    for J in 0 .. JLast loop Prod(S,C1(J),C2(J)); end loop;
  end Prod;

  function "*"(S: Scalar; Q: LChain) return LChain is
    Qt: LChain;
  begin
    Prod(S,Q,Qt);
    return Qt;
  end "*";

  procedure AddProd(S: in Scalar; V1: in LVector; V2: in out LVector) is
  begin
    if IsZero(S) then
      null;
    elsif S=SOne then
      Add(V1,V2);
    else
      for J in V1'Range loop AddProd(S,V1(J),V2(J)); end loop;
    end if;
  end AddProd;

  procedure AddProd(S: in Scalar; Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    if not IsZero(S) then
      if Q2.W>Q1.W then Q2.W := Q1.W; end if;
      if Q1.Par /= Q2.Par then raise Parity_Error; end if;
      AddProd(S,Q1.A,Q2.A);
      for J in 0 .. JLast loop AddProd(S,C1(J),C2(J)); end loop;
    end if;
  end AddProd;

  procedure Mult(Q1: in LChain; Q2: in out LChain) is
    --- pointwise multiplication
    A1: Scalar renames Q1.A;
    A2: Scalar renames Q2.A;
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);

    procedure Mult_Loop is
      J: Integer;
      Tmp: Legend;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        exit when J>JLast;
        SetZero(Tmp);
        AddProd(C1(J),C2(J),Tmp);
        AddProd(A1,C2(J),Tmp);
        AddProd(A2,C1(J),Tmp);
        Swap(Tmp,C2(J));
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Mult_Loop;

    NT: constant Natural := Reserve_Tasks(JCMax,Prod_Parallel);
  begin
    Q2.W := abs(Q1.W*(-Q2.W));
    Q2.Par := (Q1.Par+Q2.Par) mod 2;
    if NT=0 then
      Mult_Loop;
    else
      declare
        task type Mult_Task_Type is end Mult_Task_Type;
        task body Mult_Task_Type is begin Mult_Loop; end Mult_Task_Type;
        Mult_Task: array(1 .. NT) of Mult_Task_Type;
      begin
        Mult_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    Mult(A1,A2);
    if not Consistent(Q2) then raise Inconsistent_Data; end if;
  end Mult;

  procedure Prod(Q1,Q2: in LChain; Q3: in out LChain) is
    --- pointwise product
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
    C3: LGerms renames Q3.C;
    Task_Error: Boolean := False;
    PC: Protected_Counter(0);

    procedure Prod_Loop is
      J: Integer;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        exit when J>JLast;
        Prod(C1(J),C2(J),C3(J));
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    NT: constant Natural := Reserve_Tasks(JCMax,Prod_Parallel);
  begin
    Q3.W := abs(Q1.W*(-Q2.W));
    Q3.Par := (Q1.Par+Q2.Par) mod 2;
    if NT=0 then
      Prod_Loop;
    else
      declare
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    Prod(Q1.A,Q2.A,Q3.A);
    AddProd(Q1.A,C2,Q3.C);
    AddProd(Q2.A,C1,Q3.C);
    if not Consistent(Q3) then raise Inconsistent_Data; end if;
  end Prod;

  function "*"(Q1,Q2: LChain) return LChain is
    Q3: LChain;
  begin
    Prod(Q1,Q2,Q3);
    return Q3;
  end "*";

  procedure Norm(W: in Radius; Q: in LChain; S: in out Scalar) is
    C: LGerms renames Q.C;
    Tmp: Scalar;
  begin
    if not IsZero(Q.A) then raise Undefined; end if;
    if STrunc then SetZero(S); else Norm(C(JEMax),S); end if;
    for J in reverse 1 .. JCMax loop
      Mult(W,S,Tmp);
      Norm(C(J),Tmp);
      Add(Tmp,S);
    end loop;
    Mult(Two*W,S,Tmp);
    Norm(C(0),Tmp);
    Add(Tmp,S);
  end Norm;

  procedure Norm(Q: in LChain; S: in out Scalar) is
  begin
    Norm(Q.W,Q,S);
  end Norm;

  function MaxNorm(W: Radius; Q: LChain) return Radius is
    C: LGerms renames Q.C;
    R: Radius;
  begin
    if not IsZero(Q.A) then raise Undefined; end if;
    if STrunc then
      R := Zero;
    else
      R := MaxNorm(C(JEMax));
    end if;
    for J in reverse 1 .. JCMax loop
      R := W*R+MaxNorm(C(J));
    end loop;
    R := Two*W*R+MaxNorm(C(0));
    return R;
  end MaxNorm;

  function MaxNorm(Q: LChain) return Radius is
  begin
    return MaxNorm(Q.W,Q);
  end MaxNorm;

  procedure Val(Q: in LChain; T: in Scalar; Y: in out Scalar) is
  begin
    if MaxNorm(T) <= Half then
      Val(Q.C(0),T,Y);
    elsif Approx(T)<Zero then
      Val(Q,-T,Y);
      if Q.Par=1 then Neg(Y); end if;
    else
      declare
        J: constant Natural := Integer(Flt'Floor(Approx(T)+Half));
        X: Scalar;
      begin
        Diff(T,Scal(J),X);
        if J <= JLast then
          Val(Q.C(J),X,Y);
        elsif STrunc then
          SetZero(Y);
        else
          Val(Q.C(JEMax),X,Y);
          Mult(Prot_Power(One/Q.W,J-JEMax),Y);
        end if;
      end;
    end if;
  end Val;

  procedure AddCoeff(J,I: in Natural; S: in Scalar; Q: in out LChain) is
  begin
    if IsZero(S) then
      null;
    elsif J<JEmax then
      AddCoeff(I,S,Q.C(J));
    elsif Not_Strunc then
      AddCoeff(I,Ballat0(NPower(Q.W,J-JEMax)*MaxNorm(S)),Q.C(JEMax));
    end if;
  end AddCoeff;

  procedure AddBall(R: in Radius; Q: in out LChain) is
    W: Radius renames Q.W;
    C: LGerms renames Q.C;
    Fac: Radius := Half*R;
  begin
    if not (STrunc or else R=Zero) then
      AddHigher(Q.Par,R,C(0));
      for J in 1 .. JEMax loop
        Fac := Fac/W;
        AddHigher(0,Fac,C(J));
        AddHigher(1,Fac,C(J));
      end loop;
    end if;
  end AddBall;

  procedure AddBall(S: in Scalar; Q: in out LChain) is
  begin
    if Not_STrunc then AddBall(MaxNorm(S),Q); end if;
  end AddBall;

  procedure Component(J: in Integer; Q: in LChain; P: in out Legend) is
    N: constant Integer := abs(J);
    C: LGerms renames Q.C;
  begin
    if N <= JCMax then                            --- plain coeff site
      if J >= 0 then
        Copy(C(N),P);
      elsif Q.Par=0 then
        Reflect(C(N),P);
      else
        NegReflect(C(N),P);
      end if;
    elsif STrunc then
      SetZero(P);
    elsif N=JEMax then                           --- first tail
      Copy(C(JEMax),P);
    else                                         --- tail
      Prod(Prot_Power(One/Q.W,N-JEMax),C(JEMax),P);
    end if;
    if not IsZero(Q.A) then
      if Q.Par=1 then raise Inconsistent_Data; end if;
      AddCoeff(0,Q.A,P);
    end if;
  end Component;

  procedure Add_Site(J: in Integer; P: in Legend; Q: in out LChain; Check: in Boolean) is
    --- meant for J >= 0
    C: LGerms renames Q.C;
  begin
    if J<0 then
      if Check then raise Sorry; end if;
    elsif IsZero(P) then
      null;
    elsif J <= JCMax then                              --- plain coeff site
      Add(P,C(J));
      if J=0 then
        if Check and then not CheckPar(Q.Par,P) then raise Parity_Error; end if;
        ProjPar(Q.Par,C(0));
      end if;
    elsif STrunc then
      null;
    elsif J=JEMax then                                 --- first tail
      AddToErr(P,C(JEMax));
    else                                               --- tail
      AddProdToErr(Prot_Power(Q.W,J-JEMax),P,C(JEMax));
    end if;
  end Add_Site;

  procedure Add_Tail(J: in Positive; P: in Legend; Q: in out LChain) is
    --- E is assumed to be a tail starting from J
    --- not checking Center0(P)
    W: Radius renames Q.W;
    C: LGerms renames Q.C;
    Fac: Radius;
  begin
    if STrunc or else IsZero(P) then
      null;
    elsif J=JEMax then
      Add(P,C(JEMax));
    elsif J>JEMax then
      AddProd(Prot_Power(W,J-JEMax),P,C(JEMax));
    else
      Fac := One;
      for K in J .. JEMax loop
        AddProd(Fac,P,C(J));
        Fac := Fac/W;
      end loop;
    end if;
  end Add_Tail;

  procedure Reflect(Q: in out LChain) is
    C: LGerms renames Q.C;
  begin
    if Q.Par=0 then
      for J in JCRange loop Reflect(C(J)); end loop;
    else
      for J in JCRange loop NegReflect(C(J)); end loop;
    end if;
  end Reflect;

  procedure Reflect(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Copy(Q1.A,Q2.A);
    if Q1.Par=0 then
      for J in JCRange loop Reflect(C1(J),C2(J)); end loop;
    else
      for J in JCRange loop NegReflect(C1(J),C2(J)); end loop;
    end if;
    if Not_STrunc then Copy(C1(JEMax),C2(JEMax)); end if;
  end Reflect;

  function UpDown(Q: LChain; BCut: Radius := Zero) return FltVec is
    --- finds some of the bumps
    C: LGerms renames Q.C;
    N: constant Integer := 128;        --- a power of 2
    MaxCount: constant Positive := 512;
    N2: constant Flt := Flt(2*N);
    Count: Natural := 0;
    KMin: Integer := 1-N;
    Direction: Integer := 0;
    YInf,YSup,MinYAbs: Flt;
    LastYInf,LastYSup: Flt := Zero;
    V: FltVec(1..MaxCount);
    X,Y: Scalar;
  begin
    for J in reverse 0 .. JCMax loop
      if J=0 then KMin := 0; end if;
      for K in reverse KMin .. N loop
        Assign(Flt(K)/N2,X);
        Val(C(J),X,Y);
        YInf := Inf(Y);
        YSup := Sup(Y);
        if YInf*YSup>Zero then
          MinYAbs := RMax(YInf,-YSup);
          if Direction>0 then      -- was going up
            if YSup<LastYInf then  -- now going down
              Direction := -1;
              if MinYAbs>BCut then
                Count := Count+1;
                V(Count) := YSup;
              end if;
            end if;
          elsif Direction<0 then   -- was going down
            if YInf>LastYSup then  -- now going up
              Direction := 1;
              if MinYAbs>BCut then
                Count := Count+1;
                V(Count) := YInf;
              end if;
            end if;
          elsif YInf>Zero then
            Direction := 1;
          else
            Direction := -1;
          end if;
          LastYInf := YInf;
          LastYSup := YSup;
        end if;
      end loop;
    end loop;
    declare
      Par: Parity renames Q.Par;
      Last: constant Integer := Count;
      First: constant Integer := Par-Count;
      W: FltVec(First .. Last);
    begin
      if Par=0 then
        if Direction=1 then W(0) := YInf; else W(0) := YSup; end if;
      end if;
      for K in 1 .. Count loop
        W(Last-K+1) := V(K);
        if Par=0 then W(First+K-1) := V(K); else W(First+K-1) := -V(K); end if;
      end loop;
      return W;
    end;
  end UpDown;

  procedure Show1(N: in String; Q: in LChain; Hide0: in Boolean := True) is
  begin
    if not IsZero(Q) then
      Show1(N & "W ",Q.W);
      Show1(N & "P ",Q.Par);
      Show1(N & "A ",Q.A);
      if STrunc then
        Show1(N,Q.C(JCRange),Hide0);
      else
        Show1(N,Q.C,Hide0);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
    if not Consistent(Q) then raise Inconsistent_Data; end if;
  end Show1;

  procedure Put(F: in File_Type; Q: in LChain; Decimal: in Boolean := False) is
    C: LGerms renames Q.C;
  begin
    if not Consistent(Q) then raise Inconsistent_Data; end if;
    Put(F,Q.W,Decimal);
    Put(F,Q.Par);
    Put(F,JEMax);
    for J in JCRange loop
      Put(F,C(J),Decimal);
    end loop;
    if STrunc then
      PutL0(F,Decimal);
    else
      Put(F,C(JEMax),Decimal);
    end if;
    if not IsZero(Q.A) then
      Show0("Warning: nonzero Q.A not saved");
    end if;
  end Put;

  procedure Get(F: in File_Type; Q: in out LChain; Decimal: in Boolean := False) is
    JEMaxSaved: Positive := 1;
    Par: Parity := 0;
    W: Radius := Zero;
    P: Legend;
  begin
    Get(F,W,Decimal);
    Get(F,Par);
    SetZero(W,Par,Q);
    Get(F,JEMaxSaved);
    for J in 0 .. JEMaxSaved-1 loop
      Get(F,P,Decimal);
      Add_Site(J,P,Q,True);
    end loop;
    Get(F,P,Decimal);
    Add_Tail(JEMaxSaved,P,Q);
    if not Consistent(Q) then raise Inconsistent_Data; end if;
  end Get;

  procedure Write(FileName: in String; Q: in LChain; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,Q,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; Q: in out LChain; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,Q,Decimal);
    Close(F);
  end Read;

  -------------------- modes

  procedure Norm(M: in LMode; S: in out Scalar) is
  begin
    if M.LErr or else M.J=JEMax then
      Copy(SOne,S);  --- error modes have norm 1
    else
      declare
        S0,S1,Tmp: Scalar;
      begin
        Assign(Rho,S0);
        IPower(M.I,S0,S,Tmp);
        if M.J>0 then
          Assign(M.W,S0);
          IPower(M.J,S0,S1,Tmp);
          Mult(S1,S);
          Mult(Two,S);
        end if;
      end;
    end if;
  end Norm;

  procedure MaxNorm(M: in LMode; S: out Radius) is
    J: Integer renames M.J;
  begin
    if M.LErr or else J=JEMax then
      S := One;      --- error modes have norm 1
    elsif J=0 then
      S := RhoPow(M.I);
    else
      S := Two*RhoPow(M.I)*Prot_Power(M.W,J);
    end if;
  end MaxNorm;

  procedure Assign(M: in LMode; C: in out LVector) is
    --- assuming that M is fine
    I: ERange1 renames M.I;
    J: JERange renames M.J;
    P: Legend renames C(J);
    R: Radius;
  begin
    SetZero(C);
    if M.LErr or else J=JEMax then
      if J=0 then
        R := InvRhoPow(I);
      else
        R := Half*InvRhoPow(I)*Prot_Power(One/M.W,J);
      end if;
      if M.LErr then
        P.E(I) := R;
      else                       -- J=JEMax
        AddCoeff(I,BallAt0(R),P);
      end if;
    else
      AddCoeff(I,SOne,P);
    end if;
  end Assign;

  procedure Assign(M: in LMode; Q: in out LChain) is
  begin
    Q.W := M.W;
    Q.Par := M.Par;
    SetZero(Q.A);
    Assign(M,Q.C);
  end Assign;

  procedure AddProd(M: in LMode; S: in Scalar; C: in out LVector) is
    --- assuming M is fine
    J: JERange renames M.J;
    I: ERange1 renames M.I;
    P: Legend renames C(J);
    R: Radius;
  begin
    if M.LErr or else J=JEMax then
      if J=0 then
        R := InvRhoPow(I)*MaxNorm(S);
      else
        R := Half*InvRhoPow(I)*Prot_Power(One/M.W,J)*MaxNorm(S);
      end if;
      if M.LErr then
        P.E(I) := P.E(I)+R;
      else                       -- J=JEMax
        AddCoeff(I,BallAt0(R),P);
      end if;
    else
      AddCoeff(I,S,P);
    end if;
  end AddProd;

  procedure AddProd(M: in LMode; S: in Scalar; Q: in out LChain; Check: in Boolean) is
  begin
    if not IsZero(S) then
      if Check then
        if Q.Par /= M.Par then raise Parity_Error; end if;
        if Q.W /= M.W then
          if Not_STrunc and then Q.W<M.W then raise Domain_Error; end if;
          Q.W := M.W;
        end if;
      end if;
      AddProd(M,S,Q.C);
    end if;
  end AddProd;

  procedure Extract_Coeff(M: in LMode; C: in out LVector; S: in out Scalar) is
    --- assuming M.LErr=False and M.J <= JCMax
    --- assuming M and C are of the same type
    --- S encloses the coefficient
    I: ERange1 renames M.I;
    P: Legend renames C(M.J);
    PC: Poly renames P.L.all.Data;
    PE: Errs renames P.E;
  begin
    if M.Simple or else PE(I)=Zero then
      SetZero(S);
      Swap(PC(I),S);
    else
      BallAt0(PE(I),S);
      Add(PC(I),S);
      SetZero(PC(I));
      PE(I+2) := PE(I+2)+InvRho2*PE(I);
      PE(I) := Zero;
    end if;
  end Extract_Coeff;

  procedure Extract_ErrCoeff(M: in LMode; C: in out LVector; S: in out Scalar) is
    --- assuming M.LErr=False and M.J=JEMax
    --- assuming M and C are of the same type
    --- S encloses an upper bound on the norm
    I: ERange1 renames M.I;
    P: Legend renames C(JEMax);
    PC: Poly renames P.L.all.Data;
    PE: Errs renames P.E;
    R: Radius := MaxNorm(PC(I));
  begin
    if M.Simple then
      null;
    elsif PE(I) /= Zero then
      R := R+PE(I);
      PE(I+2) := PE(I+2)+InvRho2*PE(I);
      PE(I) := Zero;
    end if;
    if R=Zero then
      SetZero(S);
    else
      SetZero(PC(I));
      BallAt0(Two*RhoPow(I)*Prot_Power(M.W,JEMax)*R,S);
    end if;
  end Extract_ErrCoeff;

  procedure Extract_Legend_Error(M: in LMode; C: in out LVector; S: in out Scalar) is
    --- assuming M.LErr=True
    --- assuming M and C are of the same type
    --- S enloses an upper bound on the norm
    J: JERange renames M.J;
    I: ERange1 renames M.I;
    P: Legend renames C(J);
    PC: Poly renames P.L.all.Data;
    PE: Errs renames P.E;
    K: Integer;
    SC,SE: Radius := Zero;
  begin
    if M.Simple then
      SE := PE(I);
      PE(I) := Zero;
    else
      K := ELast0+(I mod 2);
      while K >= I loop
        SE := Rho2*SE+PE(K);
        PE(K) := Zero;
        K := K-2;
      end loop;
      K := CLast0+(I mod 2);
      while K >= I loop
        if not IsZero(PC(K)) then
          SC := Rho2*SC+MaxNorm(PC(K));
          SetZero(PC(K));
        end if;
        K := K-2;
      end loop;
    end if;
    if J=0 then
      BallAt0(RhoPow(I)*(SC+SE),S);
    else
      BallAt0(Two*RhoPow(I)*Prot_Power(M.W,J)*(SC+SE),S);
    end if;
  end Extract_Legend_Error;

  procedure Extract(M: in LMode; Q: in out LChain; S: in out Scalar; Check: in Boolean) is
    C: LGerms renames Q.C;
    --- assuming M is consistent
  begin
    if Check then
      if Q.Par /= M.Par then raise Parity_Error; end if;
      if Q.W /= M.W then
        if Not_STrunc and then Q.W<M.W then raise Domain_Error; end if;
        Q.W := M.W;
      end if;
    end if;
    if M.LErr then
      Extract_Legend_Error(M,C,S);
      --- last step, if Not_STrunc
      if M.J=JEMax and then ((M.I mod 2)=1) and then (not IsZero(Q)) then raise Sorry; end if;
    elsif M.J<JEMax then
      Extract_Coeff(M,C,S);
    else
      Extract_ErrCoeff(M,C,S);
    end if;
  end Extract;

  procedure MCheck(M: in LMode) is
    Bad: exception;
  begin
    if M.J<JEMax then
      if M.J=0 and then (M.I mod 2) /= M.Par then raise Bad; end if;
      if STrunc and then (M.LErr or (not M.Simple)) then raise Bad; end if;
    elsif STrunc then
      raise Bad;
    end if;
  exception
    when Bad =>
      Show0("!!! inconsistent mode:");
      Show(M);
      raise Inconsistent_Data;
  end MCheck;

  procedure ShowIJ(N: in String; M: in LMode; NewLine: in Boolean := True) is
  begin
    if M.Lerr then
      Show0(N & " L_Err:  J=" & Strng(M.J,3) & " I=" & Strng(M.I,3),NewLine);
    elsif M.J=JEMax then
      Show0(N & " Err_C:  J=" & Strng(M.J,3) & " I=" & Strng(M.I,3),NewLine);
    else
      Show0(N & " Coeff:  J=" & Strng(M.J,3) & " I=" & Strng(M.I,3),NewLine);
    end if;
  end ShowIJ;

  procedure ShowType(M: in LMode) is
  begin
    Show1("     W =",M.W);
    Show1("   Par =",M.Par);
    Show1("Simple = ",M.Simple);
  end ShowType;

  procedure Show(M: in LMode) is
  begin
    ShowType(M);
    Show1("  LErr = ",M.LErr);
    Show1("     J =",M.J);
    Show1("     I =",M.I);
  end Show;

  function SameType(M1,M2: LMode) return Boolean is
  begin
    return (M1.W=M2.W) and (M1.Par=M2.Par) and (M1.Simple=M2.Simple);
  end SameType;

  procedure Show(M: in LModes) is
    F: constant Positive := M'First;
    MF: LMode renames M(F);
    OK: Boolean;
  begin
    if F /= 1 then raise Inconsistent_Data; end if;
    ShowType(MF);
    ShowIJ("   1",MF);
    for K in 2 .. M'Last loop
      OK := SameType(MF,M(K));
      ShowIJ(Strng(K,4),M(K),OK);
      if not OK then Show0("!!! wrong type"); end if;
    end loop;
  end Show;

  function Compare(M1,M2: in LMode) return Integer is
  begin
    if M1.J<M2.J then return -1; end if;
    if M1.J>M2.J then return  1; end if;
    if M2.LErr and then not M1.LErr then return -1; end if;
    if M1.LErr and then not M2.LErr then return  1; end if;
    if M1.I<M2.I then return -1; end if;
    if M1.I>M2.I then return  1; end if;
    return 0;
  end Compare;

  procedure Check_Adjacent(M1,M2: in LMode) is
    Bad: exception;
  begin
    if not SameType(M1,M2) then raise Bad; end if;
    if Compare(M1,M2) >= 0 then raise Bad; end if;
  exception
    when Bad =>
      Show0("!!! inconsistent mode pair:");
      Show(M1);
      Show(M2);
      raise Inconsistent_Data;
  end Check_Adjacent;

  procedure MCheck(M: in LModes) is
    --- does not check partition property
  begin
    if M'First /= 1 then raise Inconsistent_Data; end if;
    MCheck(M(1));
    for K in 2 .. M'Last loop
      MCheck(M(K));
      Check_Adjacent(M(K-1),M(K));
    end loop;
  end MCheck;

  function Extract(J: Natural; M: LModes) return LModes is
    --- extract modes M(K) with M(K).J=J
    L: Natural := 0;
    MP: LModes := M;
  begin
    for K in M'Range loop
      if M(K).J=J then L := L+1; MP(L) := M(K); end if;
    end loop;
    return MP(1 .. L);
  end Extract;

  function Find(M: LModes; I,J: Natural) return Natural is
  begin
    for K in M'Range loop
      if M(K).J=J and then M(K).I=I then return K; end if;
    end loop;
    return 0;
  end Find;

  function Gen_Make(JCut: Positive; W: Radius; Par: Parity; Simple,Include_Errs: Boolean) return LModes is
    JMax: constant Positive := Choose(Include_Errs,JEMax,JCMax);

    function Checked_MaxDeg(J: in Natural) return Integer is
      D: Integer := MaxDeg(J);
    begin
      if D>CLast1 then D := CLast1; elsif D<-1 then D := -1; end if;
      --- last LMode will have I odd, if Not_STrunc
      if J=JEMax and then (D mod 2)=0 then D := D-1; end if;
      return D;
    end Checked_MaxDeg;

    procedure Build(M: out LModes; Count: out Natural) is
      procedure Set_Mode(J,I: in Natural; LErr: in Boolean) is
      begin
        if J>0 or else (I mod 2)=Par then
          Count := Count+1;
          if Count <= M'Last then M(Count) := (I,J,W,Par,LErr,Simple); end if;
        end if;
      end Set_Mode;
      D: Integer;
    begin
      Count := 0;
      for J in 0 .. IMin(JCut,JMax) loop
        D := Checked_MaxDeg(J);
        for I in 0 .. D loop
          Set_Mode(J,I,False);     --- coefficient (error if J=JEMax)
        end loop;
        if Include_Errs then
          Set_Mode(J,D+1,True);    --- Legend error
          Set_Mode(J,D+2,True);    --- Legend error
        end if;
      end loop;
    end Build;
    Count: Natural;
  begin
    declare
      M: LModes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: LModes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Gen_Make;

  function Make(Size: Positive; W: Radius; Par: Parity; Simple,Trunc: Boolean) return LModes is
    function MaxDeg(J: Natural) return Integer is begin return Size-J; end MaxDeg;
    function GMake is new Gen_Make (MaxDeg => MaxDeg);
    S: constant Boolean := Simple or Trunc or STrunc;
    Include_Errs: constant Boolean := not (Trunc or STrunc);
  begin
    return GMake(JEMax,W,Par,S,Include_Errs);
  end Make;

  -------------------- numerical

  procedure NumTrunc(Q: in out LChain) is
    C: LGerms renames Q.C;
  begin
    if Not_STrunc then
      Center(Q.A);
      for J in JERange loop NumTrunc(C(J)); end loop;
    end if;
  end NumTrunc;

  procedure NumTrunc(Q1: in LChain; Q2: in out LChain) is
    C1: LGerms renames Q1.C;
    C2: LGerms renames Q2.C;
  begin
    Q2.W := Q1.W;
    Q2.Par := Q1.Par;
    Center(Q1.A,Q2.A);
    for J in 0 .. JLast loop NumTrunc(C1(J),C2(J)); end loop;
  end NumTrunc;

  function NumVal(Q: LChain; T: in Flt) return Flt is
    AbsT: constant Radius := abs(T);
    J: Natural;
  begin
    if AbsT <= Half then return NumVal(Q.C(0),T); end if;
    if T >= Zero then
      J := Integer(Flt'Floor(T+Half));
      if J <= JCMax then return NumVal(Q.C(J),T-Flt(J)); end if;
      return Zero;
    elsif Q.Par=0 then
      return NumVal(Q,AbsT);
    else
      return -NumVal(Q,AbsT);
    end if;
  end NumVal;

  function V_Last(N: Positive; JCut: JCRange := JCMax) return Positive is
    --- used e.g. in NumVals and Simple_Random
  begin
    return (2*JCut+1)*N;
  end V_Last;

  procedure NumVals(N: in Positive; Q: in LChain; Y: out FltVec) is
    --- values Y(K)=NumVal(Q,T) where T=K/(2*N)
    --- assuming Y'Range is reasonable
    T: Flt;
  begin
    for K in Y'Range loop
      T := Flt(K)/Flt(2*N);
      Y(K) := NumVal(Q,T);
    end loop;
  end NumVals;

  function NumVals(N: Positive; Q: LChain; JCut: JCRange := JCMax) return FltVec is
    --- values Y(K)=NumVal(Q,T) where T=K/(2*N)
    L: constant Positive := V_Last(N,JCut);
    Y: FltVec(0 .. L);
  begin
    NumVals(N,Q,Y);
    return Y;
  end NumVals;

  procedure NumConvert(N: in Positive; Y: in FltVec; Par: in Parity; Q: in out LChain) is
    --- Y defines a function H via H(T)=Y(2*N*T)
    --- Y'Range must include 0
    --- set suitable Q.W afterwards
    WLarge: constant Radius := Two**32;
    F: Integer := -N;
    L: Integer;
    V: FltVec(-N .. Y'Last);
  begin
    SetZero(WLarge,Par,Q);
    for K in -N .. -1 loop V(K) := Flt(1-2*Par)*Y(-K); end loop;
    for K in 0 .. V'Last loop V(K) := Y(K); end loop;
    if Par=1 then V(0) := Zero; end if;
    for J in 0 .. JCMax loop
      L := F+2*N;
      exit when L>V'Last;
      NumConvert(V(F .. L),Q.C(J));
      F := L;
    end loop;
    ProjPar(Par,Q.C(0));
  end NumConvert;

  procedure NumInterpolate(Q1: in LChain; Q2: in out LChain; Fac2: in Radius) is
    --- Q2 := Fac1*Q1+Fac2*Q2, Fac1=1-Fac2
  begin
    if Fac2=Zero then
      Copy(Q1,Q2);
    elsif Fac2 /= One then
      declare
        C1: LGerms renames Q1.C;
        C2: LGerms renames Q2.C;
        Fac1: constant Radius := One-Fac2;
      begin
        if Q2.W>Q1.W then Q2.W := Q1.W; end if;
        if Q2.Par /= Q1.Par then raise Parity_Error; end if;
        Mult(Fac2,Q2.A);
        AddProd(Fac1,Q1.A,Q2.A);
        for J in 0 .. JCMax loop NumInterpolate(C1(J),C2(J),Fac2); end loop;
        if Not_Strunc then SetZero(C2(JEMax)); end if;
      end;
    end if;
  end NumInterpolate;

  function Simple_Random(Decay: Radius; Par: Parity; N: Positive) return FltVec is
    L: constant Positive := V_Last(N);
    package FC is new FltChains (VLast => L, Par => Par, N => N);
    use FC;
  begin
    return Simple_Random(Decay);
  end Simple_Random;

  procedure Simple_Random(Decay: in Radius; Par: in Parity; Q: in out LChain) is
    N: constant Positive := 64; -- subdivision
    Y: constant FltVec := Simple_Random(Decay,Par,N);
  begin
    NumConvert(N,Y,Par,Q);
  end Simple_Random;

  procedure WritePlot(FileName: in String; Par: in Parity; N: in Positive; V: in FltVec) is
    L: constant Positive := V'Last;
    package FC is new FltChains (VLast => L, Par => Par, N => N);
    use FC;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    XFac := XFactor;
    PutPlot(F,V);
    Close(F);
  end WritePlot;

  procedure PutPlot(F: in File_Type; Q: in LChain; JCut: in JCRange := JCMax) is
    DX: constant Radius := Half**6;
    P: Legend;
  begin
    for J in -JCut .. JCut loop
      Component(J,Q,P);
      PutPlot(F,P,DX,Flt(J));
    end loop;
  end PutPlot;

  procedure WritePlot(FileName: in String; Q: in LChain; JCut: in JCRange := JCMax) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    PutPlot(F,Q,JCut);
    Close(F);
  end WritePlot;

  procedure WritePlotSum(FileName: in String; Q: in LChain; JCut: in JCRange := JCMax) is
    N: constant Positive := 64;
    V: constant FltVec := NumVals(N,Q,JCut);
    L: constant Positive := V'Last;
    package FC is new FltChains (VLast => L, Par => Q.Par, N => N);
    use FC;
    U: FltVec(-L .. L);
    F: File_Type;
  begin
    XFac := XFactor;
    SumR(V,U);
    Symmetrize(U,1-Q.Par);
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    PutPlot(F,U);
    Close(F);
  end WritePlotSum;

  procedure Translate(Q1: in LChain; Q2: in out LChain; T,N: in Positive) is
    V: constant FltVec := NumVals(N,Q1,JCMax);
    L: constant Positive := V'Last;
    package FC is new FltChains (VLast => L, Par => Q1.Par, N => N);
    use FC;
    U: constant FltVec(V'Range) := Translate(V,T);
  begin
    NumConvert(N,U,Q2.Par,Q2);
  end Translate;

  ------------------------ traveling-wave related

  procedure Make_DDInvCoeffs(D: in out DData) is
    N: Integer;
  begin
    for I in 2 .. ELast1+2 loop
      N := 4*(2*I+1);
      Quot(SOne,Flt(N*(2*I-1)),D.CMinus(I));
      Quot(SOne,Flt(N*(2*I+3)),D.CPlus(I));
      Sum(D.CMinus(I),D.CPlus(I),D.CZero(I));
      D.AbsCMinus(I) := MaxNorm(D.CMinus(I));
      D.AbsCPlus(I) := MaxNorm(D.CPlus(I));
      D.AbsCZero(I) := MaxNorm(D.CZero(I));
    end loop;
  end Make_DDInvCoeffs;

  procedure DDInvHigh(D: in DData; P: in Legend; P0: in out Legend) is
    --- modes I=2,3,...
    C: Poly renames P.L.all.Data;
    C0: Poly renames P0.L.all.Data;
    CZero: DDS renames D.CZero;
    CMinus: DDS renames D.CMinus;
    CPlus: DDS renames D.CPlus;
    Tmp: Scalar;
  begin
    SetZero(P0);
    --- coeff --> coeff
    for I in 2 .. CLast1 loop
      if not IsZero(C(I)) then
        AddProd(CMinus(I),C(I),C0(I-2));
        if I<CLast0 then AddProd(CPlus(I),C(I),C0(I+2)); end if;
        SubProd(CZero(I),C(I),C0(I),Tmp);
      end if;
    end loop;
    if Not_STrunc then
      declare
        E: Errs renames P.E;
        E0: Errs renames P0.E;
        AbsCZero: DDR renames D.AbsCZero;
        AbsCMinus: DDR renames D.AbsCMinus;
        AbsCPlus: DDR renames D.AbsCPlus;
        Fac: Radius;
      begin
        --- coeff --> err
        E0(ELast0) := E0(ELast0)+AbsCPlus(ELast0-2)*MaxNorm(C(ELast0-2));
        E0(ELast1) := E0(ELast1)+AbsCPlus(ELast1-2)*MaxNorm(C(ELast1-2));
        --- low err --> err
        for I in 2 .. CLast1 loop
          if E(I) /= Zero then
            E0(I-2) := E0(I-2)+AbsCMinus(I)*E(I);
            E0(I+2) := E0(I+2)+AbsCPlus(I)*E(I);
            E0(I) := E0(I)+AbsCZero(I)*E(I);
          end if;
        end loop;
        --- high err --> err
        if E(Elast0) /= Zero then
          E0(ELast0-2) := E0(ELast0-2)+AbsCMinus(ELast0)*E(ELast0);
          Fac := AbsCMinus(ELast0+2)*InvRho2+AbsCPlus(ELast0)*Rho2+AbsCZero(ELast0);
          E0(ELast0) := E0(ELast0)+Fac*E(ELast0);
        end if;
        if E(Elast1) /= Zero then
          E0(ELast1-2) := E0(ELast1-2)+AbsCMinus(ELast1)*E(ELast1);
          Fac := AbsCMinus(ELast1+2)*InvRho2+AbsCPlus(ELast1)*Rho2+AbsCZero(ELast1);
          E0(ELast1) := E0(ELast1)+Fac*E(ELast1);
        end if;
      end;
    end if;
  end DDInvHigh;

  ---------- Scale=1

  procedure AddAALow0(C: in Poly; CL,C0,CR: in out Poly) is
    --- add A^2(S*P0) for Scale=1
    S: Scalar renames C(0);
    S00,S02,S11,S12: Scalar;
    S10: Scalar renames S02;
  begin
    Quot(S,Flt(3)/Flt(2),S00);
    Quot(S,Flt(6),S02);
    Quot(S,Flt(4),S11);
    Quot(S,Flt(12),S12);
    ---
    Add(S10,CL(0));
    Add(S11,CL(1));
    Add(S12,CL(2));
    Add(S00,C0(0));
    Sub(S02,C0(2));
    Add(S10,CR(0));
    Sub(S11,CR(1));
    Add(S12,CR(2));
  end AddAALow0;

  procedure AddAALow1(C: in Poly; CL,C0,CR: in out Poly) is
    --- add A^2(S*P1) for Scale=1
    S: Scalar renames C(1);
    S01,S03,S10,S11,S13: Scalar;
  begin
    Quot(S,Flt(5),S01);
    Quot(S,Flt(30),S03);
    Quot(S,Flt(12),S10);
    Quot(S,Flt(10),S11);
    Quot(S,Flt(60),S13);
    ---
    Sub(S10,CL(0));
    Sub(S11,CL(1));
    Add(S13,CL(3));
    Add(S01,C0(1));
    Sub(S03,C0(3));
    Add(S10,CR(0));
    Sub(S11,CR(1));
    Add(S13,CR(3));
  end AddAALow1;

  procedure PreAA1(P: in Legend; PL,P0,PR: in out Legend) is
    --- A^2 for Scale=1
    --- assuming P.E(0) and P.E(1) are zero
    --- P0 is written, PL and PR are added
    C: Poly renames P.L.all.Data;
    CL: Poly renames PL.L.all.Data;
    C0: Poly renames P0.L.all.Data;
    CR: Poly renames PR.L.all.Data;
  begin
    --- I=2,3,...
    DDInvHigh(DD,P,P0);
    Add(P0,PL);
    Add(P0,PR);
    Mult(-Two,P0);
    --- I=0,1
    if not IsZero(C(0)) then AddAALow0(C,CL,C0,CR); end if;
    if not IsZero(C(1)) then AddAALow1(C,CL,C0,CR); end if;
  end PreAA1;

  procedure AA1(V,VTmp: in out LVector; LTmp: in out Legend) is
    --- need Scale=1, V'Range=VTmp'Range
    JF: constant Integer := V'First+1;
    JL: constant Integer := V'Last-1;
    STmp: Scalar;
  begin
    Swap(V,VTmp);
    SetZero(V);
    for J in JF .. JL loop
      if not IsZero(VTmp(J)) then
        if Not_STrunc then Raise_Errs(1,VTmp(J),STmp); end if;
        PreAA1(VTmp(J),V(J-1),LTmp,V(J+1));
        Add(LTmp,V(J));
      end if;
    end loop;
  end AA1;

  ---------- Scale=2

  procedure AddAALow0(C: in Poly; CL2,CL1,C0,CR1,CR2: in out Poly) is
    --- add 4*A^2(S*P0) for Scale=2
    S: Scalar renames C(0);
    S00,S02,S10,S11,S21,S22: Scalar;
    S20: Scalar renames S02;
  begin
    Quot(S,Flt(3),S00);
    Mult(Flt(5),S00);
    Quot(S,Flt(6),S02);
    Copy(S,S10);
    Quot(S,Flt(2),S11);
    Quot(S,Flt(4),S21);
    Quot(S,Flt(12),S22);
    ---
    Add(S20,CL2(0));
    Add(S21,CL2(1));
    Add(S22,CL2(2));
    Add(S10,CL1(0));
    Add(S11,CL1(1));
    Add(S00,C0(0));
    Sub(S02,C0(2));
    Add(S10,CR1(0));
    Sub(S11,CR1(1));
    Add(S20,CR2(0));
    Sub(S21,CR2(1));
    Add(S22,CR2(2));
  end AddAALow0;

  procedure AddAALow1(C: in Poly; CL2,CL1,C0,CR1,CR2: in out Poly) is
    --- add 4*A^2(S*P1) for Scale=2
    S: Scalar renames C(1);
    S01,S03,S10,S20,S21,S23: Scalar;
  begin
    Quot(S,Flt(5),S01);
    Quot(S,Flt(30),S03);
    Quot(S,Flt(6),S10);
    Quot(S,Flt(12),S20);
    Quot(S,Flt(10),S21);
    Quot(S,Flt(60),S23);
    ---
    Sub(S20,CL2(0));
    Sub(S21,CL2(1));
    Add(S23,CL2(3));
    Sub(S10,CL1(0));
    Add(S01,C0(1));
    Sub(S03,C0(3));
    Add(S10,CR1(0));
    Add(S20,CR2(0));
    Sub(S21,CR2(1));
    Add(S23,CR2(3));
  end AddAALow1;

  procedure PreAA2(P: in Legend; PL2,PL1,P0,PR1,PR2: in out Legend) is
    --- 4*A^2 for Scale=2
    --- assuming P.E(0) and P.E(1) are zero
    --- P0 is written, PL and PR are added
    C: Poly renames P.L.all.Data;
    CL2: Poly renames PL2.L.all.Data;
    CL1: Poly renames PL1.L.all.Data;
    C0: Poly renames P0.L.all.Data;
    CR1: Poly renames PR1.L.all.Data;
    CR2: Poly renames PR2.L.all.Data;
  begin
    --- I=2,3,...
    DDInvHigh(DD,P,P0);
    Add(P0,PL2);
    Add(P0,PR2);
    Mult(-Two,P0);
    --- I=0,1
    if not IsZero(C(0)) then AddAALow0(C,CL2,CL1,C0,CR1,CR2); end if;
    if not IsZero(C(1)) then AddAALow1(C,CL2,CL1,C0,CR1,CR2); end if;
  end PreAA2;

  procedure AA2(V,VTmp: in out LVector; LTmp: in out Legend) is
    --- need Scale=2, V'Range=VTmp'Range
    JF: constant Integer := V'First+2;
    JL: constant Integer := V'Last-2;
    STmp: Scalar;
  begin
    Swap(V,VTmp);
    SetZero(V);
    for J in JF .. JL loop
      if not IsZero(VTmp(J)) then
        if Not_STrunc then Raise_Errs(1,VTmp(J),STmp); end if;
        PreAA2(VTmp(J),V(J-2),V(J-1),LTmp,V(J+1),V(J+2));
        Add(LTmp,V(J));
      end if;
    end loop;
    for J in JF .. JL loop Mult(Quarter,V(J)); end loop;
  end AA2;

  ---------- Scale=1 or Scale=2

  procedure AA(V,VTmp: in out LVector; LTmp: in out Legend) is
  begin
    if Scale=1 then
      AA1(V,VTmp,LTmp);
    elsif Scale=2 then
      AA2(V,VTmp,LTmp);
    else
      raise Not_Implemented;
    end if;
  end AA;

  ---------- Low --> All

  function LA_JF(K: Positive) return Integer is
    --- index of lowest germ needed to compute Q.C(0..)
    N: constant Positive := Scale*K;
  begin
    if N>JCMax then raise Not_Implemented; end if;
    return -N;
  end LA_JF;

  function LA_JL(K: Positive) return Positive is
    --- index of highest germ affected by Q.C(..JCMax)
  begin
    if STrunc then return JCMax+(Scale*K)/2; end if;
    return JCMax+Scale*K;
  end LA_JL;

  procedure LA_Expand(Q: in LChain; V: in out LVector) is
  begin
    for J in V'First .. JCMax loop Component(J,Q,V(J)); end loop;
    for J in JEMax .. V'Last loop SetZero(V(J)); end loop;
  end LA_Expand;

  procedure LA_UnExpand(N: in Positive; V: in LVector; Q: in out LChain; LTmp: in out Legend) is
    --- SetZero(W,Par,Q) beforehand
    C: LGerms renames Q.C;
    Fac: Radius := One;
  begin
    ProjPar(Q.Par,V(0),C(0));
    for J in 1 .. JCMax loop Copy(V(J),C(J)); end loop;
    if Not_STrunc then
      ToErr(V(JEMax),C(JEMax));
      for J in JEMax+1 .. JEMax+Scale*N loop
        ToErr(V(J),LTmp);
        Fac := Q.W*Fac;
        AddProd(Fac,LTmp,C(JEMax));
      end loop;
    end if;
  end LA_UnExpand;

  procedure LA_AAPower(K: in Positive; Q: in out LChain; W: in Radius) is
    Par: constant Parity := Q.Par;
    JF: constant Integer := LA_JF(K);
    JL: constant Integer := LA_JL(K);
    LTmp: Legend;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
  begin
    LA_Expand(Q,V);
    for N in 1 .. K loop AA(V,VTmp,LTmp); end loop;
    SetZero(W,Par,Q);
    LA_UnExpand(K,V,Q,LTmp);
  end LA_AAPower;

  procedure LA_SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar) is
    --- AA+mu*AA^2+mu^2*AA^3+...
    Par: constant Parity := Q.Par;
    JF: constant Integer := LA_JF(K);
    JL: constant Integer := LA_JL(K);
    Fac: Scalar := SOne;
    LTmp: Legend;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
    Qt: LChain;
  begin
    LA_Expand(Q,V);
    SetZero(W,Par,Q);
    for N in 1 .. K loop
      AA(V,VTmp,LTmp);
      SetZero(W,Par,Qt);
      LA_UnExpand(N,V,Qt,LTmp);
      AddProd(Fac,Qt,Q);
      Mult(Mu,Fac);
    end loop;
  end LA_SumAAPowers;

  ---------- High --> Low

  function HL_JF(K: Positive) return Positive is
    --- index of lowest germ affected by Q.C(JEMax)
  begin
    return JEMax-Scale*K;
  end HL_JF;

  function HL_JL(K: Positive) return Positive is
    --- index of highest germ affecting Q.C(JCMax)
  begin
    return JCMax+Scale*K;
  end HL_JL;

  procedure HL_Expand(Q: in LChain; V: in out LVector) is
  begin
    for J in V'First .. JCMax loop SetZero(V(J)); end loop;
    for J in JEMax .. V'Last loop Component(J,Q,V(J)); end loop;
  end HL_Expand;

  procedure HL_UnExpand(V: in LVector; Q: in out LChain) is
    --- SetZero(W,Par,Q) beforehand
    C: LGerms renames Q.C;
  begin
    for J in V'First .. JCMax loop Copy(V(J),C(J)); end loop;
  end HL_UnExpand;

  procedure HL_AAPower(K: in Positive; Q: in out LChain; W: in Radius) is
    Par: constant Parity := Q.Par;
    JF: constant Integer := HL_JF(K);
    JL: constant Integer := HL_JL(K);
    LTmp: Legend;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
  begin
    HL_Expand(Q,V);
    for N in 1 .. K loop AA(V,VTmp,LTmp); end loop;
    SetZero(W,Par,Q);
    HL_UnExpand(V,Q);
  end HL_AAPower;

  procedure HL_SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar) is
    --- AA+mu*AA^2+mu^2*AA^3+...
    Par: constant Parity := Q.Par;
    JF: constant Integer := HL_JF(K);
    JL: constant Integer := HL_JL(K);
    Fac: Scalar := SOne;
    LTmp: Legend;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
    Qt: LChain;
  begin
    HL_Expand(Q,V);
    SetZero(W,Par,Q);
    for N in 1 .. K loop
      AA(V,VTmp,LTmp);
      SetZero(W,Par,Qt);
      HL_UnExpand(V,Qt);
      AddProd(Fac,Qt,Q);
      Mult(Mu,Fac);
    end loop;
  end HL_SumAAPowers;

  ---------- High --> High

  function HH_JF(K: Positive) return Positive is
    --- index of lowest germ affecting Q.C(JEMax)
  begin
    return JEMax-Scale*K;
  end HH_JF;

  function HH_JL(K: Positive) return Positive is
    --- index of highest germ affecting Q.C(JEMax)
  begin
    return JEMax+Scale*K;
  end HH_JL;

  procedure HH_Expand(W: in Radius; P: in Legend; V: in out LVector) is
    Fac: Radius := One;
  begin
    Copy(P,V(JEMax));
    for J in JEMax+1 .. V'Last loop
      Fac := Fac/W;
      Prod(Fac,P,V(J));
    end loop;
    Fac := One;
    for J in reverse V'First .. JEMax-1 loop
      Fac := Fac*W;
      Prod(Fac,P,V(J));
    end loop;
  end HH_Expand;

  procedure HH_Unexpand(V: in LVector; P: in out Legend) is
  begin
    Copy(V(JEMax),P);
  end HH_Unexpand;

  procedure HH_AAPower(K: in Positive; P: in out Legend; W: in Radius) is
    JF: constant Integer := HH_JF(K);
    JL: constant Integer := HH_JL(K);
    LTmp: Legend renames P;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
  begin
    HH_Expand(W,P,V);
    for N in 1 .. K loop AA(V,VTmp,LTmp); end loop;
    HH_UnExpand(V,P);
  end HH_AAPower;

  procedure HH_SumAAPowers(K: in Positive; P: in out Legend; W: in Radius; Mu: in Scalar) is
    --- AA+mu*AA^2+mu^2*AA^3+...
    JF: constant Integer := HH_JF(K);
    JL: constant Integer := HH_JL(K);
    Fac: Scalar := SOne;
    Pt: Legend;
    LTmp: Legend renames Pt;
    V,VTmp: LVector(JF-Scale .. JL+Scale);
  begin
    HH_Expand(W,P,V);
    SetZero(P);
    for N in 1 .. K loop
      AA(V,VTmp,LTmp);
      HH_UnExpand(V,Pt);
      AddProd(Fac,Pt,P);
      Mult(Mu,Fac);
    end loop;
  end HH_SumAAPowers;

  ---------- Combined

  procedure AAPower(K: in Positive; Q: in out LChain; W: in Radius) is
  begin
    if IsFinite(Q) then
      LA_AAPower(K,Q,W);
    else
      declare
        Pt: Legend;
        Qt: LChain;
      begin
        Copy(Q,Qt);
        HL_AAPower(K,Qt,W);   -- High --> Low
        Copy(Q.C(JEMax),Pt);
        HH_AAPower(K,Pt,W);   -- High --> High
        LA_AAPower(K,Q,W);    -- Low  --> All
        Add(Qt,Q);
        Add(Pt,Q.C(JEMax));
      end;
    end if;
  end AAPower;

  procedure SumAAPowers(K: in Positive; Q: in out LChain; W: in Radius; Mu: in Scalar) is
  begin
    if IsFinite(Q) then
      LA_SumAAPowers(K,Q,W,Mu);
    else
      declare
        Pt: Legend;
        Qt: LChain;
      begin
        Copy(Q,Qt);
        HL_SumAAPowers(K,Qt,W,Mu);  -- High --> Low
        Copy(Q.C(JEMax),Pt);
        HH_SumAAPowers(K,Pt,W,Mu);  -- High --> High
        LA_SumAAPowers(K,Q,W,Mu);   -- Low  --> All
        Add(Qt,Q);
        Add(Pt,Q.C(JEMax));
      end;
    end if;
  end SumAAPowers;

  ---------- Numeric

  procedure NumAA(Par: in Parity; N: in Positive; V1: in FltVec; V2: out FltVec) is
    L: constant Positive := V1'Last;
    package FC is new FltChains (VLast => L, Par => Par, N => N);
    use FC;
  begin
    if V2'Last /= L then raise Dimension_Error; end if;
    AA(Scale,V1,V2);
  end NumAA;

begin

  XFactor := One/Flt(Scale); --- for plotting
  Make_DDInvCoeffs(DD);

end Legends.Chain;
