with Globals, Strings;
use Globals, Strings;

pragma Elaborate_All (Globals,Strings);

package body Legends is

  use SV;

  ------------------ Poly

  procedure CP(C1: in Poly; C2: in out Poly) is
  begin
    for I in CRange1 loop Copy(C1(I),C2(I)); end loop;
  end CP;

  function Deg(C: Poly) return Natural is
  begin
    for I in reverse 1 .. CLast1 loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 0;
  end Deg;

  function LastC(C: Poly) return Integer is
  begin
    for I in reverse CRange1 loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return -1;
  end LastC;

  procedure Mult(S: in Scalar; C: in out Poly) is
  begin
    for I in CRange1 loop Mult(S,C(I)); end loop;
  end Mult;

  procedure Prod(S: in Scalar; C1: in Poly; C2: in out Poly) is
  begin
    for I in CRange1 loop Prod(S,C1(I),C2(I)); end loop;
  end Prod;

  procedure AddProd(S: in Scalar; C1: in Poly; C2: in out Poly) is
    Tmp: Scalar;
  begin
    for I in CRange1 loop AddProd(S,C1(I),C2(I),Tmp); end loop;
  end AddProd;

  procedure SubProd(S: in Scalar; C1: in Poly; C2: in out Poly) is
    Tmp: Scalar;
  begin
    for I in CRange1 loop SubProd(S,C1(I),C2(I),Tmp); end loop;
  end SubProd;

  function MaxNorm(C: Poly) return Radius is
    D: constant Natural := Deg(C);
    R: Radius := MaxNorm(C(D));
  begin
    for I in reverse 0 .. D-1 loop
      R := Rho*R+MaxNorm(C(I));
    end loop;
    return R;
  end MaxNorm;

  procedure Val(C: in Poly; X: in Scalar; Y: in out Scalar) is
    --- using Bonnet's formula
    D: constant Natural := Deg(C);
    X2: constant Scalar := Two*X;
    I: Integer := 2;
    P0,P1,P2: Scalar;
  begin
    Copy(C(0),Y);
    AddProd(C(1),X2,Y);
    if D>1 then
      Copy(SOne,P2);            -- P_{I-2}
      Copy(X2,P1);              -- P_{I-1}
      loop
        Prod(Flt(2*I-1),P1,P0);
        Mult(X2,P0);
        AddProd(Flt(1-I),P2,P0);
        Div(Flt(I),P0);          -- P_I
        AddProd(C(I),P0,Y);
        exit when I=D;
        Copy(P1,P2);
        Copy(P0,P1);
        I := I+1;
      end loop;
    end if;
  end Val;

  procedure AddToErr(C1: in Poly; C2: in out Poly) is
    Tmp: Scalar;
  begin
    for I in CRange1 loop ToErr(C1(I),Tmp); Add(Tmp,C2(I)); end loop;
  end AddToErr;

  procedure AddProdToErr(R: in Flt; C1: in Poly; C2: in out Poly) is
    Tmp: Scalar;
  begin
    for I in CRange1 loop ToErr(C1(I),Tmp); AddProd(R,Tmp,C2(I)); end loop;
  end AddProdToErr;

  procedure L2Prod(C1,C2: in Poly; S: in out Scalar) is
    T,Tmp: Scalar;
  begin
    SetZero(S);
    for I in reverse CRange1 loop
      Prod(C1(I),C2(I),T);
      Div(Flt(I)+Half,T,Tmp);
      Add(T,S);
    end loop;
  end L2Prod;

  -------------------- Errs

  function EffLast(E: Errs) return Natural is
  begin
    for I in reverse 1 .. ELast1 loop
      if E(I) /= Zero then return I; end if;
    end loop;
    return 0;
  end EffLast;

  function LastE(E: Errs) return Integer is
  begin
    if Not_STrunc then
      for I in reverse ERange1 loop
        if E(I) /= Zero then return I; end if;
      end loop;
    end if;
    return -1;
  end LastE;

  function FirstE(E: Errs) return Natural is
    --- assuming Not_STrunc
  begin
    for I in ERange0 loop
      if E(I) /= Zero then return I; end if;
    end loop;
    return ELast1;
  end FirstE;

  function "="(E1,E2: Errs) return Boolean is
  begin
    for I in reverse 1 .. ELast1 loop
      if E1(I) /= E2(I) then return False; end if;
    end loop;
    return E1(0)=E2(0);
  end "=";

  function IsZero(E: Errs) return Boolean is
  begin
    for I in reverse 1 .. ELast1 loop
      if E(I) /= Zero then return False; end if;
    end loop;
    return E(0)=Zero;
  end IsZero;

  procedure Swap(E1,E2: in out Errs) is
    R: Radius;
  begin
    for I in ERange1 loop
      R := E1(I); E1(I) := E2(I); E2(I) := R;
    end loop;
  end Swap;

  procedure Add(C1: in Poly; E2: in out Errs) is
  begin
    for I in CRange1 loop
      if not IsZero(C1(I)) then E2(I) := E2(I)+MaxNorm(C1(I)); end if;
    end loop;
  end Add;

  procedure Add(E1: in Errs; E2: in out Errs) is
  begin
    for I in ERange1 loop E2(I) := E2(I)+E1(I); end loop;
  end Add;

  procedure Sum(E1,E2: in Errs; E3: out Errs) is
  begin
    for I in ERange1 loop E3(I) := E1(I)+E2(I); end loop;
  end Sum;

  procedure Mult(R: in Radius; E: in out Errs) is
  begin
    for I in ERange1 loop E(I) := R*E(I); end loop;
  end Mult;

  procedure Prod(R: in Radius; E1: in Errs; E2: out Errs) is
  begin
    for I in ERange1 loop E2(I) := R*E1(I); end loop;
  end Prod;

  procedure AddProd(R: in Radius; C1: in Poly; E2: in out Errs) is
  begin
    for I in CRange1 loop
      if not IsZero(C1(I)) then E2(I) := E2(I)+R*MaxNorm(C1(I)); end if;
    end loop;
  end AddProd;

  procedure AddProd(R: in Radius; E1: in Errs; E2: in out Errs) is
  begin
    for I in ERange1 loop E2(I) := E2(I)+R*E1(I); end loop;
  end AddProd;

  procedure Put(F: in File_Type; E: in Errs; Decimal: in Boolean := False) is
    D: Natural;
  begin
    if STrunc then
      Put(F,0);
      Put(F,Zero,Decimal);
    else
      D := EffLast(E);
      Put(F,D);
      for I in 0 .. D loop Put(F,E(I),Decimal); end loop;
    end if;
  end Put;

  procedure Get(F: in File_Type; E: out Errs; Decimal: in Boolean := False) is
    D: Natural := 0;
    S,Fac: Radius := One;
  begin
    Get(F,D);
    E := (others => Zero);
    for I in 0 .. D loop
      if I <= ELast1 then
        Get(F,E(I),Decimal);
      else
        Get(F,S,Decimal);
        Fac := Fac*Rho;
        if (I mod 2)=0 then
          E(ELast0) := E(ELast0)+Fac*S;
        else
          E(ELast1) := E(ELast1)+Fac*S;
        end if;
      end if;
    end loop;
  end Get;

  function MaxNorm0(E: Errs; L: Natural) return Radius is
    --- MaxNorm of even part
    I: Natural := 2*(L/2);
    R: Radius := E(I);
  begin
    while I /= 0 loop
      I := I-2;
      R := Rho2*R+E(I);
    end loop;
    return R;
  end MaxNorm0;

  function MaxNorm1(E: Errs; L: Natural) return Radius is
    --- MaxNorm of odd part, if multiplied by Rho
    I: Integer;
    R: Radius := Zero;
  begin
    if (L mod 2)=0 then I := L-1; else I := L; end if;
    while I > 0 loop
      R := Rho2*R+E(I);
      I := I-2;
    end loop;
    return R;
  end MaxNorm1;

  function MaxNorm(E: Errs) return Radius is
    N: Radius := E(ELast1);
  begin
    for I in reverse ERange0 loop N := Rho*N+E(I); end loop;
    return N;
  end MaxNorm;

  function Coeff(N: Natural; E: Errs) return Radius is
    --- upper bound
    I: Integer := N mod 2;
    C: Radius := Zero;
  begin
    while I <= N loop
      C := InvRho2*C;
      if I <= ELast1 then C := C+E(I); end if;
      I := I+2;
    end loop;
    return C;
  end Coeff;

  function RhoPower(I: Integer) return Radius is
    --- upper bound
    K: Integer;
    X: Flt;
    Y: Flt := One;
  begin
    if I /= 0 then
      if I>0 then
        K := I;
        X := Rho;
      else
        K := -I;
        X := InvRho;
      end if;
      loop
        if (K rem 2)=1 then
          Y := X*Y;
        end if;
        K := K/2;
        exit when K=0;
        X := X*X;
      end loop;
    end if;
    return Y;
  end RhoPower;

  procedure AddCoeff(I: in Natural; R: in Radius; E: in out Errs) is
  begin
    if I <= ELast1 then
      E(I) := E(I)+R;
    elsif (I mod 2)=0 then
      E(ELast0) := E(ELast0)+RhoPower(I-ELast0)*R;
    else
      E(ELast1) := E(ELast1)+RhoPower(I-ELast1)*R;
    end if;
  end AddCoeff;

  function L2ProdPar(P: Parity; E1,E2: Errs) return Radius is
    --- upper bound
    I: Integer := ELast0+P;
    R,F1,F2: Radius := Zero;
  begin
    while I >= 0 loop
      F1 := E1(I)+InvRho2*F1;
      R := R+(E1(I)*F2+F1*E2(I))/(Flt(I)+Half);
      F2 := E2(I)+InvRho2*F2;
      I := I-2;
    end loop;
    return R;
  end L2ProdPar;

  -------------------- Legend

  function Info(Dummy: Legend) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return LInfo;
  end Info;

  function "="(P1,P2: Legend) return Boolean is
    --- equality as sets
  begin
    if (P1.L.all.Data /= P2.L.all.Data) then return False; end if;
    return STrunc or else P1.E=P2.E;
  end "=";

  function IsZero(P: Legend) return Boolean is
  begin
    return IsZero(P.L.all.Data) and then (STrunc or else IsZero(P.E));
  end IsZero;

  procedure SetZero(P: in out Legend) is
  begin
    SetZero(P.L.all.Data);
    if Not_STrunc then P.E := (others => Zero); end if;
  end SetZero;

  procedure Copy(P1: in Legend; P2: in out Legend) is
  begin
    CP(P1.L.all.Data,P2.L.all.Data);
    if Not_STrunc then P2.E := P1.E; end if;
  end Copy;

  procedure Swap(P1,P2: in out Legend) is
    use RP.Alloc;
    L1: constant LD_Pointer := P1.L;
  begin
    P1.L := P2.L; P2.L := L1;
    if Not_STrunc then Swap(P1.E,P2.E); end if;
  end Swap;

  procedure Neg(P: in out Legend) is
  begin
    Neg(P.L.all.Data);
  end Neg;

  procedure Neg(P1: in Legend; P2: in out Legend) is
  begin
    Neg(P1.L.all.Data,P2.L.all.Data);
    if Not_Strunc then P2.E := P1.E; end if;
  end Neg;

  function "-"(P: Legend) return Legend is
    Q: Legend;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(P1: in Legend; P2: in out Legend) is
  begin
    Add(P1.L.all.Data,P2.L.all.Data);
    if Not_Strunc then Add(P1.E,P2.E); end if;
  end Add;

  procedure Sum(P1,P2: in Legend; P3: in out Legend) is
  begin
    Sum(P1.L.all.Data,P2.L.all.Data,P3.L.all.Data);
    if Not_Strunc then Sum(P1.E,P2.E,P3.E); end if;
  end Sum;

  function "+"(P1,P2: Legend) return Legend is
    P3: Legend;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in Legend; P2: in out Legend) is
  begin
    Sub(P1.L.all.Data,P2.L.all.Data);
    if Not_Strunc then Add(P1.E,P2.E); end if;
  end Sub;

  procedure Diff(P1,P2: in Legend; P3: in out Legend) is
  begin
    Diff(P1.L.all.Data,P2.L.all.Data,P3.L.all.Data);
    if Not_Strunc then Sum(P1.E,P2.E,P3.E); end if;
  end Diff;

  function "-"(P1,P2: Legend) return Legend is
    P3: Legend;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P: in out Legend) is
  begin
    if R=Zero then
      SetZero(P);
    else
      Mult(Scal(R),P.L.all.Data);
      if Not_STrunc then Mult(abs(R),P.E); end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in Legend; P2: in out Legend) is
  begin
    if R=Zero then
      SetZero(P2);
    else
      Prod(Scal(R),P1.L.all.Data,P2.L.all.Data);
      if Not_STrunc then Prod(abs(R),P1.E,P2.E); end if;
    end if;
  end Prod;

  procedure AddProd(R: in Flt; P1: in Legend; P2: in out Legend) is
  begin
    if R /= Zero then
      AddProd(Scal(R),P1.L.all.Data,P2.L.all.Data);
      if Not_STrunc then AddProd(abs(R),P1.E,P2.E); end if;
    end if;
  end AddProd;

  procedure Mult(S: in Scalar; P: in out Legend) is
    R: constant Radius := MaxNorm(S);
  begin
    if R=Zero then
      SetZero(P);
    else
      Mult(S,P.L.all.Data);
      if Not_STrunc then Mult(R,P.E); end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in Legend; P2: in out Legend) is
    R: constant Radius := MaxNorm(S);
  begin
    if R=Zero then
      SetZero(P2);
    else
      Prod(S,P1.L.all.Data,P2.L.all.Data);
      if Not_STrunc then Prod(R,P1.E,P2.E); end if;
    end if;
  end Prod;

  procedure AddProd(S: in Scalar; P1: in Legend; P2: in out Legend) is
    R: constant Radius := MaxNorm(S);
  begin
    if R /= Zero then
      AddProd(S,P1.L.all.Data,P2.L.all.Data);
      if Not_STrunc then AddProd(R,P1.E,P2.E); end if;
    end if;
  end AddProd;

  procedure SubProd(S: in Scalar; P1: in Legend; P2: in out Legend) is
    R: constant Radius := MaxNorm(S);
  begin
    if R /= Zero then
      SubProd(S,P1.L.all.Data,P2.L.all.Data);
      if Not_STrunc then AddProd(R,P1.E,P2.E); end if;
    end if;
  end SubProd;

  procedure Coeff(K,L,M: in Natural; C,Tmp1,Tmp2: in out Scalar) is
    --- check beforehand that K,L,M are sane
    S: constant Integer := (K+L+M)/2;
  begin
    Prod(Flt(2*M+1),Binom(S-K),C);
    Prod(Binom(S-L),Binom(S-M),Tmp2);
    Prod(C,Tmp2,Tmp1);                --- (2*M+1)*Binom(S-K)*Binom(S-L)*Binom(S-M)
    Prod(Flt(2*S+1),Binom(S),Tmp2);   --- (2*S+1)*Binom(S)
    Quot(Tmp1,Tmp2,C);
  end Coeff;

  procedure AddProdCC(L1,L2: in Natural; C1,C2: in Poly; C3: in out Poly; E3L0,E3L1: in out Radius) is
    I3,I3Max,I3Cut: Integer;
    C,Coe,RemCoe,Tmp1,Tmp2: Scalar;
  begin
    for I1 in 0 .. L1 loop
      if not IsZero(C1(I1)) then
        for I2 in 0 .. L2 loop
          if not IsZero(C2(I2)) then
            Prod(C1(I1),C2(I2),C);
            Copy(SOne,RemCoe);
            I3Max := I1+I2;
            I3Cut := IMin(I3Max,CLast1)-2;
            I3 := abs(I1-I2);
            loop
              Coeff(I1,I2,I3,Coe,Tmp1,Tmp2);
              AddProd(Coe,C,C3(I3),Tmp1);
              Sub(Coe,RemCoe);
              exit when I3>I3Cut; -- for speed ...
              I3 := I3+2;
            end loop;
            if Not_STrunc and then I3<I3Max then
              if (I3Max mod 2)=0 then
                E3L0 := E3L0+Sup(RemCoe)*MaxNorm(C)*RhoPow(I3Max-ELast0); -- not optimal
              else
                E3L1 := E3L1+Sup(RemCoe)*MaxNorm(C)*RhoPow(I3Max-ELast1); -- not optimal
              end if;
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end AddProdCC;

  procedure AddProdCE(L1,L2: in Natural; C1: in Poly; E2: in Errs; E3: in out Errs) is
    F2: constant Natural := FirstE(E2);
    I3: Integer;
    B1: Radius;
  begin
    for I1 in 0 .. L1 loop
      B1 := MaxNorm(C1(I1));
      if B1 /= Zero then
        for I2 in F2 .. L2 loop
          I3 := I2-I1;
          if I3<0 then I3 := (I3 mod 2); end if;
          E3(I3) := E3(I3)+RhoPow(I1+I2-I3)*B1*E2(I2); -- not optimal
        end loop;
      end if;
    end loop;
  end AddProdCE;

  procedure Mult(P1: in Legend; P2,Tmp: in out Legend) is
  begin
    SetZero(Tmp);
    AddProd(P1,P2,Tmp);
    Swap(Tmp,P2);
  end Mult;

  procedure Mult(P1: in Legend; P2: in out Legend) is
    Tmp: Legend;
  begin
    SetZero(Tmp);
    AddProd(P1,P2,Tmp);
    Swap(Tmp,P2);
  end Mult;

  procedure Prod(P1,P2: in Legend; P3: in out Legend) is
  begin
    SetZero(P3);
    AddProd(P1,P2,P3);
  end Prod;

  function "*"(P1,P2: Legend) return Legend is
    P3: Legend;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";

  procedure AddProd(P1,P2: in Legend; P3: in out Legend) is
    C2: Poly renames P2.L.all.Data;
    LC2: constant Integer := LastC(C2);                --  -1 if C2=0
    E2: Errs renames P2.E;
    LE2: constant Integer := LastE(E2);                --  -1 if E2=0
  begin
    if (LC2<0) and then (LE2<0) then return; end if;   -- P2=0
    declare
      C1: Poly renames P1.L.all.Data;
      LC1: constant Integer := LastC(C1);              --  -1 if C1=0
      E1: Errs renames P1.E;
      LE1: constant Integer := LastE(E1);              --  -1 if E1=0
    begin
      if (LC1<0) and then (LE1<0) then return; end if; -- P1=0
      if (LC1>=0) and then (LC2>=0) then
        declare
          C3: Poly renames P3.L.all.Data;
          E3L0: Radius renames P3.E(ELast0);
          E3L1: Radius renames P3.E(ELast1);
        begin
          AddProdCC(LC1,LC2,C1,C2,C3,E3L0,E3L1);       -- coeff*coeff
        end;
        if (LE1<0) and then (LE2<0) then return; end if;
      end if;
      if (LC1>=0) and then (LE2>=0) then               -- coeff*err
        AddProdCE(LC1,LE2,C1,E2,P3.E);
      end if;
      if (LE1>=0) and then (LC2>=0) then               -- err*coeff
        AddProdCE(LC2,LE1,C2,E1,P3.E);
      end if;
      if (LE1>=0) and then (LE2>=0) then               -- err*err
        declare
          N10: constant Radius := MaxNorm0(E1,LE1);
          N11: constant Radius := MaxNorm1(E1,LE1);
          N20: constant Radius := MaxNorm0(E2,LE2);
          N21: constant Radius := MaxNorm1(E2,LE2);
          E3: Errs renames P3.E;
        begin
          E3(0) := E3(0)+N10*N20+N11*N21*Rho2;
          E3(1) := E3(1)+N10*N21+N11*N20;
        end;
      end if;
    end;
  end AddProd;

  procedure SubProd(P1,P2: in Legend; P3: in out Legend) is
  begin
    Neg(P3);
    AddProd(P1,P2,P3);
    Neg(P3);
  end SubProd;

  function ErrNorm(P: Legend) return Radius is
  begin
    if STrunc then return Zero; end if;
    declare
      E: Errs renames P.E;
      R: Radius := E(ELast1);
    begin
      for I in reverse ERange0 loop R := Rho*R+E(I); end loop;
      return R;
    end;
  end ErrNorm;

  procedure Norm(P: in Legend; S: in out Scalar) is
    C: Poly renames P.L.all.Data;
    D: constant Natural := Deg(C);
    E: constant Radius := ErrNorm(P);
    Tmp: Scalar;
  begin
    Norm(C(D),S);
    for I in reverse 0 .. D-1 loop
      Mult(Rho,S,Tmp);
      Norm(C(I),Tmp);
      Add(Tmp,S);
    end loop;
    if E /= Zero then
      BallAt0(E,Tmp);
      Add(Tmp,S);
    end if;
  end Norm;

  function MaxNorm(P: Legend) return Radius is
  begin
    return MaxNorm(P.L.all.Data)+ErrNorm(P);
  end MaxNorm;

  procedure Show1(N: in String; P: in Legend; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    C: Poly renames P.L.all.Data;
    DC: constant Natural := Deg(C);
    AllZero: Boolean;
  begin
    if not IsZero(P) then
      AllZero := True;
      for I in 0 .. DC loop
        if not IsZero(C(I)) then
          Show1(N & "C " & Strng(I,W) & " ",C(I));
          AllZero := False;
        end if;
      end loop;
      if not Hide0 then
        if AllZero then Show0(N & "C 0"); end if;
        Show0;
      end if;
      if Not_STrunc then
        AllZero := True;
        declare
          E: Errs renames P.E;
          DE: constant Natural := EffLast(E);
        begin
          for I in 0 .. DE loop
            if E(I) /= Zero then
              Show1(N & "E " & Strng(I,W) & " ",E(I));
              AllZero := False;
            end if;
          end loop;
          if not Hide0 then
            if AllZero then Show0(N & "E 0"); end if;
            Show0;
          end if;
        end;
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in Legend; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    DC: constant Natural := IMax(Deg(C1),Deg(C2));
    AllZero: Boolean := True;
  begin
    if not (IsZero(P1) and IsZero(P2)) then
      AllZero := True;
      for I in 0 .. DC loop
        if not (IsZero(C1(I)) and then IsZero(C2(I))) then
          Show2(N & "C " & Strng(I,W) & " ",C1(I),C2(I));
          AllZero := False;
        end if;
      end loop;
      if not Hide0 then
        if AllZero then Show0(N & "C 0 0"); end if;
        Show0;
      end if;
      if Not_STrunc then
        AllZero := True;
        declare
          E1: Errs renames P1.E;
          E2: Errs renames P2.E;
          DE: constant Natural := IMax(EffLast(E1),EffLast(E2));
        begin
          for I in 0 ..DE loop
            if (E1(I) /= Zero) or (E2(I) /= Zero) then
              Show2(N & "E " & Strng(I,W) & " ",E1(I),E2(I));
              AllZero := False;
            end if;
          end loop;
          if not Hide0 then
            if AllZero then Show0(N & "E 0 0"); end if;
            Show0;
          end if;
        end;
      end if;
    elsif not Hide0 then
      Show0(N & "0 0");
    end if;
  end Show2;

  procedure PutL0(F: in File_Type; Decimal: in Boolean := False) is
  begin
    Put(F,Rho,Decimal);
    Put(F,0);
    Put(F,SZero,Decimal);
    Put(F,0);
    Put(F,Zero,Decimal);
  end PutL0;

  procedure Put(F: in File_Type; P: in Legend; Decimal: in Boolean := False) is
    C: Poly renames P.L.all.Data;
    D: constant Natural := Deg(C);
  begin
    Put(F,Rho,Decimal);
    Put(F,D);
    for I in 0 .. D loop Put(F,C(I),Decimal); end loop;
    Put(F,P.E,Decimal);
  end Put;

  procedure Get(F: in File_Type; P: in out Legend; Decimal: in Boolean := False) is
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    D: Natural := 0;
    Fac: Radius := One;
    R,R0,R1: Radius := Zero;
    S: Scalar;
  begin
    Get(F,R,Decimal);
    Get(F,D);
    SetZero(C);
    for I in 0 .. D loop
      if I <= CLast1 then
        Get(F,C(I),Decimal);
      else
        Get(F,S,Decimal);
        if (I mod 2)=0 then
          R0 := R0+Fac*MaxNorm(S);
        else
          R1 := R1+Fac*MaxNorm(S);
        end if;
        Fac := Fac*Rho;
      end if;
    end loop;
    Get(F,E,Decimal);
    if R0 /= Zero then E(ELast0) := E(ELast0)+R0; end if;
    if R1 /= Zero then E(ELast1) := E(ELast1)+R1; end if;
    if R>Rho and then (not IsPoly(P)) then raise Domain_Error; end if;
  end Get;

  procedure Write(FileName: in String; P: in Legend; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Legend; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  function IsReal(P: Legend) return Boolean is
  begin
    if LInfo.IsComplex then
      if not IsReal(P.L.all.Data) then return False; end if;
      if STrunc then return True; end if;
      if not IsZero(P.E) then raise Not_Certain; end if;
    end if;
    return True;
  end IsReal;

  procedure Real_Part(P: in out Legend) is
  begin
    Real_Part(P.L.all.Data);
  end Real_Part;

  procedure Imag_Part(P: in out Legend) is
  begin
    if LInfo.IsComplex then
      Imag_Part(P.L.all.Data);
    else
      SetZero(P);
    end if;
  end Imag_Part;

  procedure Adjoint(P: in out Legend) is
  begin
    if LInfo.IsComplex then Adjoint(P.L.all.Data); end if;
  end Adjoint;

  procedure Adjoint(P1: in Legend; P2: in out Legend) is
  begin
    Adjoint(P1.L.all.Data,P2.L.all.Data);
    if Not_STrunc then P2.E := P1.E; end if;
  end Adjoint;

  function IsPoly(P: Legend) return Boolean is
  begin
    if not STrunc then
      declare
        E: Errs renames P.E;
      begin
        for I in ERange1 loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsPoly;

  procedure ToErr(P: in out Legend) is
    C: Poly renames P.L.all.Data;
  begin
    for I in CRange1 loop ToErr(C(I)); end loop;
  end ToErr;

  procedure ToErr(P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    for I in CRange1 loop ToErr(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end ToErr;

  function Center0(P: Legend) return Boolean is
    C: Poly renames P.L.all.Data;
  begin
    for I in CRange0 loop
      if not Center0(C(I)) then return False; end if;
    end loop;
    return Center0(C(CLast1));
  end Center0;

  procedure Union(P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    E1: Errs renames P1.E;
    E2: Errs renames P2.E;
  begin
    for I in CRange1 loop Union(C1(I),C2(I)); end loop;
    if Not_STrunc then
      for I in ERange1 loop E2(I) := RMax(E1(I),E2(I)); end loop;
    end if;
  end Union;

  procedure LegVal(N: in Natural; X: in Scalar; Y: in out Scalar) is
    --- value of Legendre_N at X
    --- using Bonnet's formula
  begin
    if N=0 then
      Copy(SOne,Y);
    else
      Prod(Two,X,Y);
      if N>1 then
        declare
          X2: constant Scalar := Y;
          I: Integer := 2;
          P1,P2: Scalar;
        begin
          Copy(SOne,P2);           -- P_{I-2}
          Copy(X2,P1);             -- P_{I-1}
          loop
            Prod(Flt(2*I-1),P1,Y);
            Mult(X2,Y);
            AddProd(Flt(1-I),P2,Y);
            Div(Flt(I),Y);         -- P_I
            exit when I=N;
            Copy(P1,P2);
            Copy(Y,P1);
            I := I+1;
          end loop;
        end;
      end if;
    end if;
  end LegVal;

  procedure ValLeft(P: in Legend; Y: in out Scalar) is
    --- value at -1/2
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    I: Integer := CLast1;
    R: Flt := Zero;
  begin
    if Not_Strunc then
      for K in reverse ERange1 loop R := R+E(K); end loop;
    end if;
    BallAt0(R,Y);
    loop
      Sub(C(I),Y);
      I := I-1;
      Add(C(I),Y);
      exit when I=0;
      I := I-1;
    end loop;
  end ValLeft;

  procedure ValRight(P: in Legend; Y: in out Scalar) is
    --- value at 1/2
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    R: Flt := Zero;
  begin
    if Not_Strunc then
      for K in reverse ERange1 loop R := R+E(K); end loop;
    end if;
    BallAt0(R,Y);
    for I in reverse CRange1 loop Add(C(I),Y); end loop;
  end ValRight;

  procedure Val(P: in Legend; X: in Scalar; Y: in out Scalar) is
  begin
    if X=SHalf then
      ValRight(P,Y);
    elsif X=SNegHalf then
      ValLeft(P,Y);
    else
      Val(P.L.all.Data,X,Y);
      if not IsPoly(P) then
        declare
          R: constant Radius := Two*MaxNorm(X);
          E: Errs renames P.E;
          V: Radius := E(ELast1);
          S: Scalar;
        begin
          if R>Rho then
            Show2("Rho R:",Rho,R);
            raise Domain_Error;
          end if;
          for I in reverse ERange0 loop V := R*V+E(I); end loop;
          BallAt0(V,S);
          Add(S,Y);
        end;
      end if;
    end if;
  end Val;

  function Val(P: Legend; X: Scalar) return Scalar is
    Y: Scalar;
  begin
    Val(P,X,Y);
    return Y;
  end Val;

  procedure Coeff(N: in Natural; P: in Legend; S: in out Scalar) is
    C: Poly renames P.L.all.Data;
  begin
    if IsPoly(P) then
      if N <= CLast1 then Copy(C(N),S); else SetZero(S); end if;
    else
      BallAt0(Coeff(N,P.E),S);
      if N <= CLast1 then Add(C(N),S); end if;
    end if;
  end Coeff;

  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out Legend) is
    C: Poly renames P.L.all.Data;
  begin
    if I <= CLast1 then
      Add(S,C(I));
    elsif STrunc then
      null;
    else
      AddCoeff(I,MaxNorm(S),P.E);
    end if;
  end AddCoeff;

  procedure AddHigher(I: in ERange1; R: in Radius; P: in out Legend) is
    --- add higher error with norm R
  begin
    if not STrunc then P.E(I) := P.E(I)+InvRhoPow(I)*R; end if;
  end AddHigher;

  procedure Add(P1: in Legend; E2: in out Errs) is
  begin
    if Not_STrunc then
      Add(P1.L.all.Data,E2);
      Add(P1.E,E2);
    end if;
  end Add;

  procedure AddProd(R: in Radius; P1: in Legend; E2: in out Errs) is
  begin
    if not (STrunc or else R=Zero) then
      AddProd(R,P1.L.all.Data,E2);
      AddProd(R,P1.E,E2);
    end if;
  end AddProd;

  procedure AddToErr(P1: in Legend; P2: in out Legend) is
  begin
    if Not_Strunc then
      AddToErr(P1.L.all.Data,P2.L.all.Data);
      Add(P1.E,P2.E);
    end if;
  end AddToErr;

  procedure AddProdToErr(R: in Flt; P1: in Legend; P2: in out Legend) is
  begin
    if not (STrunc or else R=Zero) then
      AddProdToErr(R,P1.L.all.Data,P2.L.all.Data);
      AddProd(R,P1.E,P2.E);
    end if;
  end AddProdToErr;

  procedure L2Prod(P1,P2: in Legend; S: in out Scalar) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    if IsPoly(P1) and then IsPoly(P2) then
      L2Prod(C1,C2,S);
    else
      SetZero(S);
      declare
        E1: Errs renames P1.E;
        E2: Errs renames P2.E;
        I: Integer;
        F1,F2: Radius;
        CE1,CE2,T,Tmp: Scalar;
      begin
        for P in 0 .. 1 loop
          BallAt0(L2ProdPar(P,E1,E2),T);
          Add(T,S);
          F1 := Zero;
          F2 := Zero;
          I := P;
          while I <= CLast0 loop
            F1 := InvRho2*F1+E1(I);
            BallAt0(F1,CE1);
            Add(C1(I),CE1);
            F2 := InvRho2*F2+E2(I);
            BallAt0(F2,CE2);
            Add(C2(I),CE2);
            Prod(CE1,CE2,T);
            Div(Flt(I)+Half,T,Tmp);
            Add(T,S);
            I := I+2;
          end loop;
        end loop;
      end;
    end if;
  end L2Prod;

  procedure ProjHigh(D: in CRange1; P: in out Legend) is
    --- project onto modes above D
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
  begin
    if Not_STrunc then
      for I in 0 .. D loop
        if E(I) /= Zero then
          E(I+2) := E(I+2)+InvRho2*E(I);
          E(I) := Zero;
        end if;
      end loop;
    end if;
    for I in 0 .. D loop SetZero(C(I)); end loop;
  end ProjHigh;

  procedure Raise_Errs(D: in CRange1; P: in out Legend; Tmp: in out Scalar) is
    --- raise errors above D
    --- assuming Not_STrunc
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
  begin
    for I in 0 .. D loop
      if E(I) /= Zero then
        BallAt0(E(I),Tmp);
        Add(Tmp,C(I));
        E(I+2) := E(I+2)+InvRho2*E(I);
        E(I) := Zero;
      end if;
    end loop;
  end Raise_Errs;

  function CheckPar(Par: Parity; P: Legend) return Boolean is
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    I: Integer := CLast1-Par;
  begin
    while I >= 0 loop
      if not IsZero(C(I)) then return False; end if;
      I := I-2;
    end loop;
    if Not_Strunc then
      I := ELast1-Par;
      while I >= 0 loop
        if E(I) /= Zero then return False; end if;
        I := I-2;
      end loop;
    end if;
    return True;
  end CheckPar;

  procedure ProjPar(Par: in Parity; P: in out Legend) is
    C: Poly renames P.L.all.Data;
    E: Errs renames P.E;
    I: Integer := CLast1-Par;
  begin
    while I >= 0 loop SetZero(C(I)); I := I-2; end loop;
    if Not_Strunc then
      I := ELast1-Par;
      while I >= 0 loop E(I) := Zero; I := I-2; end loop;
    end if;
  end ProjPar;

  procedure ProjPar(Par: in Parity; P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    E1: Errs renames P1.E;
    E2: Errs renames P2.E;
    Keep: Boolean := (Par=0);
  begin
    for I in CRange1 loop
      if Keep then Copy(C1(I),C2(I)); else SetZero(C2(I)); end if;
      Keep := not Keep;
    end loop;
    if Not_STrunc then
      Keep := (Par=0);
      for I in ERange1 loop
        if Keep then E2(I) := E1(I); else E2(I) := Zero; end if;
        Keep := not Keep;
      end loop;
    end if;
  end ProjPar;

  procedure Reflect(P: in out Legend) is
    C: Poly renames P.L.all.Data;
    I: Integer := CLast1;
  begin
    loop Neg(C(I)); exit when I=1; I := I-2; end loop;
  end Reflect;

  procedure Reflect(P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    Odd: Boolean := False;
  begin
    for I in CRange1 loop
      if Odd then Neg(C1(I),C2(I)); else Copy(C1(I),C2(I)); end if;
      Odd := not Odd;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Reflect;

  procedure NegReflect(P: in out Legend) is
    C: Poly renames P.L.all.Data;
    I: Integer := CLast0;
  begin
    loop Neg(C(I)); exit when I=0; I := I-2; end loop;
  end NegReflect;

  procedure NegReflect(P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
    Even: Boolean := True;
  begin
    for I in CRange1 loop
      if Even then Neg(C1(I),C2(I)); else Copy(C1(I),C2(I)); end if;
      Even := not Even;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end NegReflect;

  -------------------- numeric, testing

  procedure Check_Coeff_Sums is
    KMax: constant Integer := 3*CLast1/4;
    M,MMax: Integer;
    R: Flt := Zero;
    C,CSum,Tmp1,Tmp2: Scalar;
  begin
    Show1("checking only up to ",KMax);
    for K in 0 .. KMax loop
      for L in 0 .. K loop
        SetZero(CSum);
        MMax := K+L;
        M := K-L;
        loop
          Coeff(K,L,M,C,Tmp1,Tmp2);
          Add(C,CSum);
          exit when M=MMax;
          M := M+2;
        end loop;
        Diff(SOne,CSum,Tmp1);
        R := RMax(R,MaxNorm(Tmp1));
      end loop;
    end loop;
    Show1("Err =",R);
  end Check_Coeff_Sums;

  procedure NumTrunc(P: in out Legend) is
    C: Poly renames P.L.all.Data;
  begin
    if Not_STrunc then
      for I in CRange1 loop Center(C(I)); end loop;
      P.E := (others => Zero);
    end if;
  end NumTrunc;

  procedure NumTrunc(P1: in Legend; P2: in out Legend) is
    C1: Poly renames P1.L.all.Data;
    C2: Poly renames P2.L.all.Data;
  begin
    for I in CRange1 loop Center(C1(I),C2(I)); end loop;
    P2.E := (others => Zero);
  end NumTrunc;

  function NumVal(P: Legend; X: Flt) return Flt is
    Y: Scalar;
  begin
    Val(P.L.all.Data,Scal(X),Y);
    return Approx(Y);
  end NumVal;

  procedure NumVals(C: in Poly; V: out FltVec) is
    F: constant Integer := V'First; --- corresponds to X=-1/2
    L: constant Integer := V'Last;  --- corresponds to X=1/2
    X: Flt;
    Y: Scalar;
  begin
    for N in F .. L loop
      X := Flt(N-F)/Flt(L-F)-Half;
      Val(C,Scal(X),Y);
      V(N) := Approx(Y);
    end loop;
  end NumVals;

  function Simpson(V: FltVec) return Flt is
    F: constant Integer := V'First; --- corresponds to X=-1/2
    L: constant Integer := V'Last;  --- corresponds to X=1/2
    D: constant Integer := L-F;
    N: Integer := F+2;
    R: Flt := V(F)+Four*V(F+1);
  begin
    if (D mod 2)=1 then raise Not_Implemented; end if;
    loop
      R := R+Two*V(N)+Four*V(N+1);
      N := N+2;
      exit when N=L;
    end loop;
    return (R+V(L))/Flt(3*D);
  end Simpson;

  procedure NumConvert(V: in FltVec; P: in out Legend) is
    --- V are values of a function H on [-1/2,1/2]
    use Flt_EF;
    F: constant Integer := V'First; --- corresponds to X=-1/2
    L: constant Integer := V'Last;  --- corresponds to X=1/2
    Fac,S: Flt;
    W: FltVec(F..L);
    U: Poly;
  begin
    SetZero(P);
    SetZero(U);
    for I in CRange1 loop
      Fac := Sqrt(Flt(I)+Half);
      Assign(Fac,U(I));               --- normalized Legendre U
      NumVals(U,W);
      SetZero(U(I));
      for M in F .. L loop W(M) := W(M)*V(M); end loop; --- U*V
      S := Simpson(W);                                  --- integral of U*V
      AddCoeff(I,Scal(Fac*Two*S),P);
    end loop;
    NumTrunc(P);
  end NumConvert;

  procedure NumInterpolate(P1: in Legend; P2: in out Legend; Fac2: in Radius) is
    --- P2 := Fac1*P1+Fac2*P2, Fac1=1-Fac2
  begin
    if Fac2=Zero then
      Copy(P1,P2);
    elsif Fac2 /= One then
      declare
        Fac1: constant Radius := One-Fac2;
        C1: Poly renames P1.L.all.Data;
        C2: Poly renames P2.L.all.Data;
        Tmp: Scalar;
      begin
        for I in CRange1 loop
          Mult(Fac2,C2(I),Tmp);
          AddProd(Fac1,C1(I),C2(I),Tmp);
        end loop;
      end;
    end if;
    NumTrunc(P2);
  end NumInterpolate;

  procedure Simple_Random(R: in Radius; P: in out Legend) is
    C: Poly renames P.L.all.Data;
    Ratio: constant Radius := Quarter;
    Fac: Radius := R;
  begin
    for I in CRange1 loop
      Simple_Random(C(I));
      Mult(Fac,C(I));
      Fac := Ratio*Fac;
    end loop;
    NumTrunc(P);
  end Simple_Random;

  function GuessRho(P: Legend) return Radius is
    use Flt_EF;
    DMin: constant Integer := CLast/5;
    DMax: constant Integer := (4*CLast)/5;
    C: Poly renames P.L.all.Data;
    N: Integer := 0;
    A2: Flt;
    A1: Flt := MaxNorm(C(DMin));
    A0: Flt := MaxNorm(C(DMin+1));
    Q: Flt := One;
  begin
    for I in DMin+2 .. DMax loop
      A2 := A1;
      A1 := A0;
      A0 := MaxNorm(C(I));
      if (A0 /= Zero) and then (A2 /= Zero) then
        Q := Q*(A2/A0);
        N := N+2;
      end if;
    end loop;
    return Exp(Log(Q)/Flt(N));
  end GuessRho;

  procedure PutPlot(F: in File_Type; P: in Legend; DX: in Radius; Centre: in Flt := Zero) is
    use FLt_IO;
    Digs: constant Positive := 4;
    Small: constant Radius := 1.0E-4;
    X: Flt := NegHalf;
    Y: Flt;
  begin
    if DX<Small then raise Sorry; end if;
    while X<Half loop
      Put(F,XFactor*(Centre+X),3,Digs,4);
      Y := NumVal(P,X);
      Put(F,YFactor*Y,3,Digs,4);
      New_Line(F);
      X := X+DX;
    end loop;
  end PutPlot;

begin

  if Rho<One then raise Sorry with "Rho needs to be at least 1"; end if;

end Legends;
