with Ada.Text_IO, Logicals, Scalars, Ints, Rationals, Flts, Flt_Vectors;
use Ada.Text_IO, Logicals, Scalars, Ints, Rationals, Flts;
with Pooling, Vectors, Binoms;

pragma Elaborate_All (Ada.Text_IO,Logicals,Scalars,Ints,Rationals,Flts,Flt_Vectors);
pragma Elaborate_All (Pooling,Vectors,Binoms);

generic

  CLast: in Positive;   --- defines CLast0 and CLast1
  Rho: in Radius;       --- needs to be at least 1

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure IPower(I: in Integer; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Proper_Rounding is <>;

  -------------------- end_include scalars

  with package SV is new Vectors (Component => Scalar);

package Legends is

  pragma Elaborate_Body;

  CLast0: constant Positive := 2*(CLast/2);     --- even part
  CLast1: constant Positive := CLast0+1;        --- odd part
  subtype CRange0 is Integer range 0 .. CLast0;
  subtype CRange1 is Integer range 0 .. CLast1;
  subtype Poly is SV.Vector(CRange1);

  procedure CP(C1: in Poly; C2: in out Poly);
  package RP is new Pooling (Item => Poly);
  subtype Pooled_Poly is RP.Controlled_LD_Pointer;

  ELast0: constant Positive := CLast0+2;
  ELast1: constant Positive := CLast1+2;
  subtype ERange0 is Integer range 0 .. ELast0;
  subtype ERange1 is Integer range 0 .. ELast1;
  subtype Errs is RadVec(ERange1);

  type Legend is new Pooled_Poly with
    --- L: LD_Pointer := null;   -- L.all.Data(I) are the coefficients
    record
      E: Errs;                   -- higher order errors, fixed parity, normalized like coeffs
    end record;

  function Info(Dummy: Legend) return Scalar_Info;
  function "="(P1,P2: Legend) return Boolean;
  function IsZero(P: Legend) return Boolean;
  procedure SetZero(P: in out Legend);
  procedure Copy(P1: in Legend; P2: in out Legend);
  procedure Swap(P1,P2: in out Legend);
  procedure Neg(P: in out Legend);
  procedure Neg(P1: in Legend; P2: in out Legend);
  function "-"(P: Legend) return Legend;
  procedure Add(P1: in Legend; P2: in out Legend);
  procedure Sum(P1,P2: in Legend; P3: in out Legend);
  function "+"(P1,P2: Legend) return Legend;
  procedure Sub(P1: in Legend; P2: in out Legend);
  procedure Diff(P1,P2: in Legend; P3: in out Legend);
  function "-"(P1,P2: Legend) return Legend;
  procedure Mult(R: in Flt; P: in out Legend);
  procedure Prod(R: in Flt; P1: in Legend; P2: in out Legend);
  procedure AddProd(R: in Flt; P1: in Legend; P2: in out Legend);
  procedure Mult(S: in Scalar; P: in out Legend);
  procedure Prod(S: in Scalar; P1: in Legend; P2: in out Legend);
  procedure AddProd(S: in Scalar; P1: in Legend; P2: in out Legend);
  procedure SubProd(S: in Scalar; P1: in Legend; P2: in out Legend);
  procedure Mult(P1: in Legend; P2,Tmp: in out Legend);
  procedure Mult(P1: in Legend; P2: in out Legend);
  procedure Prod(P1,P2: in Legend; P3: in out Legend);
  function "*"(P1,P2: Legend) return Legend;
  procedure AddProd(P1,P2: in Legend; P3: in out Legend);
  procedure SubProd(P1,P2: in Legend; P3: in out Legend);

  procedure Norm(P: in Legend; S: in out Scalar);
  function MaxNorm(P: Legend) return Radius;
  procedure Show1(N: in String; P: in Legend; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Legend; Hide0: in Boolean := True);
  procedure PutL0(F: in File_Type; Decimal: in Boolean := False);
  procedure Put(F: in File_Type; P: in Legend; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Legend; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Legend; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Legend; Decimal: in Boolean := False);

  function IsReal(P: Legend) return Boolean;
  procedure Real_Part(P: in out Legend);
  procedure Imag_Part(P: in out Legend);
  procedure Adjoint(P: in out Legend);
  procedure Adjoint(P1: in Legend; P2: in out Legend);
  function IsPoly(P: Legend) return Boolean;

  procedure ToErr(P: in out Legend);
  procedure ToErr(P1: in Legend; P2: in out Legend);
  function Center0(P: Legend) return Boolean;
  procedure Union(P1: in Legend; P2: in out Legend);
  procedure LegVal(N: in Natural; X: in Scalar; Y: in out Scalar);
  procedure ValLeft(P: in Legend; Y: in out Scalar);
  procedure ValRight(P: in Legend; Y: in out Scalar);
  procedure Val(P: in Legend; X: in Scalar; Y: in out Scalar);
  function Val(P: Legend; X: Scalar) return Scalar;
  procedure Coeff(N: in Natural; P: in Legend; S: in out Scalar);
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out Legend);
  procedure AddHigher(I: in ERange1; R: in Radius; P: in out Legend);
  procedure Add(P1: in Legend; E2: in out Errs);
  procedure AddProd(R: in Radius; P1: in Legend; E2: in out Errs);
  procedure AddToErr(P1: in Legend; P2: in out Legend);
  procedure AddProdToErr(R: in Flt; P1: in Legend; P2: in out Legend);
  procedure L2Prod(P1,P2: in Legend; S: in out Scalar);
  procedure ProjHigh(D: in CRange1; P: in out Legend);
  procedure Raise_Errs(D: in CRange1; P: in out Legend; Tmp: in out Scalar);
  function CheckPar(Par: Parity; P: Legend) return Boolean;
  procedure ProjPar(Par: in Parity; P: in out Legend);
  procedure ProjPar(Par: in Parity; P1: in Legend; P2: in out Legend);
  procedure Reflect(P: in out Legend);
  procedure Reflect(P1: in Legend; P2: in out Legend);
  procedure NegReflect(P: in out Legend);
  procedure NegReflect(P1: in Legend; P2: in out Legend);

  --- numeric, testing
  subtype FltVec is Flt_Vectors.Vector;
  procedure Check_Coeff_Sums;
  procedure NumTrunc(P: in out Legend);
  procedure NumTrunc(P1: in Legend; P2: in out Legend);
  function NumVal(P: Legend; X: Flt) return Flt;
  procedure NumVals(C: in Poly; V: out FltVec);
  procedure NumConvert(V: in FltVec; P: in out Legend);
  procedure NumInterpolate(P1: in Legend; P2: in out Legend; Fac2: in Radius);
  procedure Simple_Random(R: in Radius; P: in out Legend);
  function GuessRho(P: Legend) return Radius;

  --- plotting
  XFactor: Radius := One;
  YFactor: Radius := One;
  procedure PutPlot(F: in File_Type; P: in Legend; DX: in Radius; Centre: in Flt := Zero);

  --- rounded up
  InvRho:    constant Radius := One/Rho;
  RhoPow:    constant RadVec := PowersOf(Rho,2*ELast1);
  InvRhoPow: constant RadVec := PowersOf(InvRho,2*ELast1);

  function RhoPower(I: Integer) return Radius;

private

  function Deg(C: Poly) return Natural;
  function LastC(C: Poly) return Integer;
  procedure Mult(S: in Scalar; C: in out Poly);
  procedure Prod(S: in Scalar; C1: in Poly; C2: in out Poly);
  procedure AddProd(S: in Scalar; C1: in Poly; C2: in out Poly);
  procedure SubProd(S: in Scalar; C1: in Poly; C2: in out Poly);
  function MaxNorm(C: Poly) return Radius;
  procedure Val(C: in Poly; X: in Scalar; Y: in out Scalar);
  procedure AddToErr(C1: in Poly; C2: in out Poly);
  procedure AddProdToErr(R: in Flt; C1: in Poly; C2: in out Poly);
  procedure L2Prod(C1,C2: in Poly; S: in out Scalar);

  function EffLast(E: Errs) return Natural;
  function LastE(E: Errs) return Integer;
  function FirstE(E: Errs) return Natural;
  function "="(E1,E2: Errs) return Boolean;
  function IsZero(E: Errs) return Boolean;
  procedure Swap(E1,E2: in out Errs);
  procedure Add(C1: in Poly; E2: in out Errs);
  procedure Add(E1: in Errs; E2: in out Errs);
  procedure Sum(E1,E2: in Errs; E3: out Errs);
  procedure Mult(R: in Radius; E: in out Errs);
  procedure Prod(R: in Radius; E1: in Errs; E2: out Errs);
  procedure AddProd(R: in Radius; C1: in Poly; E2: in out Errs);
  procedure AddProd(R: in Radius; E1: in Errs; E2: in out Errs);
  procedure Put(F: in File_Type; E: in Errs; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; E: out Errs; Decimal: in Boolean := False);
  function MaxNorm0(E: Errs; L: Natural) return Radius;
  function MaxNorm1(E: Errs; L: Natural) return Radius;
  function MaxNorm(E: Errs) return Radius;
  function Coeff(N: Natural; E: Errs) return Radius;
  procedure AddCoeff(I: in Natural; R: in Radius; E: in out Errs);
  function L2ProdPar(P: Parity; E1,E2: Errs) return Radius;

  procedure Coeff(K,L,M: in Natural; C,Tmp1,Tmp2: in out Scalar);
  procedure AddProdCC(L1,L2: in Natural; C1,C2: in Poly; C3: in out Poly; E3L0,E3L1: in out Radius);
  procedure AddProdCE(L1,L2: in Natural; C1: in Poly; E2: in Errs; E3: in out Errs);
  function ErrNorm(P: Legend) return Radius;

  function Simpson(V: FltVec) return Flt;

  SZero:    constant Scalar := Scal(Zero);
  SOne:     constant Scalar := Scal(One);
  SHalf:    constant Scalar := Scal(Half);
  SNegHalf: constant Scalar := Scal(NegHalf);
  SInfo:    constant Scalar_Info := Info(SZero);
  STrunc:   constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean := not STrunc;

  LInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                  IsStandard => SInfo.IsStandard,
                                  IsNumeric  => SInfo.IsNumeric,
                                  IsComplex  => SInfo.IsComplex,
                                  NReps      => SInfo.NReps*(CLast1+1),
                                  RMode      => SInfo.RMode);

  Rho2:    constant Radius := Rho*Rho;
  InvRho2: constant Radius := InvRho*InvRho;

  package LB is new Binoms(DMax => CLast1, Scalar => Scalar);
  use LB;
  Binom: constant Binom_Type := Make_Binom;

end Legends;
