with Ints,Flts;
use Ints,Flts;

pragma Elaborate_All (Ints,Flts);

package Params is

  type Param is
    record
      Prefix: String(1..4);    -- name
      Scale: Positive;         -- inverse of particle spacing
      Mu: Flt;                 -- quadratic coupling
      Model: Positive;         -- power of the potential
      Symm: Parity;            -- 0 means even in time, 1 odd in time
      Rho: Radius;             -- Legend domain
      Decay: Radius;           -- decay of Chain
      CLast: Positive;         -- size of Legend
      JCMax: Positive;         -- half size of chain
      AAPow: Positive;         -- max power of AA
      Mat_PackSize: Positive;  -- pack size for matrix used in Contr
      Contr_PackSize: Natural; -- used in proof (use 0 for purely numeric solution)
    end record;

  E1PZ: constant Param := (Prefix => "E1PZ", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(1)/Flt(4),
                           Model => 3,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 24,
                           JCMax => 12,
                           AAPow => 1,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E2NZ: constant Param := (Prefix => "E2NZ", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 3,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 24,
                           JCMax => 12,
                           AAPow => 1,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E10A: constant Param := (Prefix => "E10A", --- even, 1 bump, mu=0
                           Scale => 1,
                           Mu => Zero,
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 24,
                           JCMax => 12,
                           AAPow => 1,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E1PA: constant Param := (Prefix => "E1PA", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(1)/Flt(2),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 3,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E1PB: constant Param := (Prefix => "E1PB", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(3)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 64,
                           JCMax => 20,
                           AAPow => 8,
                           Mat_PackSize => 32,
                           Contr_Packsize => 64);

  E1PC: constant Param := (Prefix => "E1PC", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(63)/Flt(64),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 64,
                           JCMax => 22,
                           AAPow => 15,
                           Mat_PackSize => 32,
                           Contr_Packsize => 0);

  E1PD: constant Param := (Prefix => "E1PD", --- even, 1 bump, mu>0
                           Scale => 1,
                           Mu => Flt(7)/Flt(8),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => One+Half**3,
                           CLast => 16,
                           JCMax => 16,
                           AAPow => 8,
                           Mat_PackSize => 16,
                           Contr_Packsize => 0);

  E1NA: constant Param := (Prefix => "E1NA", --- even, 1 bump, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 24,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E1NB: constant Param := (Prefix => "E1NB", --- even, 1 bump, mu close to -1
                           Scale => 1,
                           Mu => Flt(-63)/Flt(64),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => One+Half**3,
                           CLast => 64,
                           JCMax => 20,
                           AAPow => 3,
                           Mat_PackSize => 16,
                           Contr_Packsize => 0); -- no Wr near 1

  E1NC: constant Param := (Prefix => "E1NC", --- even, 1 bump, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 24,
                           JCMax => 24,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E1ND: constant Param := (Prefix => "E1ND", --- even, 1 bump, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Three,
                           CLast => 24,
                           JCMax => 12,
                           AAPow => 3,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32); -- no Wr near 1

  E2NA: constant Param := (Prefix => "E2NA", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E2NB: constant Param := (Prefix => "E2NB", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(256),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 1,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E2NC: constant Param := (Prefix => "E2NC", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-3)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 32,
                           JCMax => 14,
                           AAPow => 8,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E3NA: constant Param := (Prefix => "E3NA", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E3NB: constant Param := (Prefix => "E3NB", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Four,
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E4NA: constant Param := (Prefix => "E4NA", --- even, 4 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 32,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 24,
                           Contr_Packsize => 32);

  E4NB: constant Param := (Prefix => "E4NB", --- even, 4 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Three,
                           CLast => 32,
                           JCMax => 14,
                           AAPow => 2,
                           Mat_PackSize => 24,
                           Contr_Packsize => 32);

  E5NA: constant Param := (Prefix => "E5NA", --- even, 3 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 64,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E5NB: constant Param := (Prefix => "E5NB", --- even, 5 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 64,
                           JCMax => 16,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E5NC: constant Param := (Prefix => "E5NC", --- even, 5 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 64,
                           JCMax => 12,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  E5ND: constant Param := (Prefix => "E5ND", --- even, 5 bumps, mu<0
                           Scale => 1,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 0,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 64,
                           JCMax => 16,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);


  O20A: constant Param := (Prefix => "O20A",  --- odd, 2 bumps, mu close to 0
                           Scale => 1,        --- WR2 = 1.00000000000773531: proof very unlikely
                           Mu => Flt(-1)/Flt(1024),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 32,
                           JCMax => 5,
                           AAPow => 1,
                           Mat_PackSize => 32,
                           Contr_Packsize => 0);

  O2NA: constant Param := (Prefix => "O2NA", --- odd, 2 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 3,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  O2NB: constant Param := (Prefix => "O2NB",  --- odd, 2 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-3)/Flt(4),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(5)/Flt(4),
                           CLast => 32,
                           JCMax => 20,
                           AAPow => 8,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  O2NC: constant Param := (Prefix => "O2NC", --- odd, 2 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(4),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Two,
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  O2ND: constant Param := (Prefix => "O2ND", --- odd, 2 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(8),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(7)/Flt(4),
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 2,
                           Mat_PackSize => 16,
                           Contr_Packsize => 32);

  O2NE: constant Param := (Prefix => "O2NE", --- odd, 2 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(64),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 1,
                           Mat_PackSize => 16,
                           Contr_Packsize => 0);

  O4NA: constant Param := (Prefix => "O4NA", --- odd, 4 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 3,
                           Mat_PackSize => 24,
                           Contr_Packsize => 32);

  O4NB: constant Param := (Prefix => "O4NB", --- odd, 4 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 32,
                           JCMax => 32,
                           AAPow => 3,
                           Mat_PackSize => 24,
                           Contr_Packsize => 32);

  O6NA: constant Param := (Prefix => "O6NA", --- odd, 6 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(9)/Flt(8),
                           CLast => 92,
                           JCMax => 32,
                           AAPow => 3,
                           Mat_PackSize => 32,
                           Contr_Packsize => 64);

  O6NB: constant Param := (Prefix => "O6NB", --- odd, 6 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 16,
                           JCMax => 20,
                           AAPow => 2,
                           Mat_PackSize => 32,
                           Contr_Packsize => 0);

  O6NC: constant Param := (Prefix => "O6NC", --- odd, 6 bumps, mu<0
                           Scale => 2,
                           Mu => Flt(-1)/Flt(2),
                           Model => 4,
                           Symm => 1,
                           Rho => One+Half**4,
                           Decay => Flt(3)/Flt(2),
                           CLast => 24,
                           JCMax => 24,
                           AAPow => 2,
                           Mat_PackSize => 24,
                           Contr_Packsize => 0);



  type Param_List is array(Positive range <>) of Param;

  PList: constant Param_List(1..27) := (E1PZ,E2NZ,E10A,E1PA,E1PB,E1PC,E1PD,E1NA,E1NB,E1NC,E2NA,E2NB,E2NC,E3NA,
                                       E3NB,E4NA,E4NB,E5NA,E5NB,E5NC,E5ND,O2ND,O4NA,O4NB,E1ND,O6NA,O2NA);

  procedure Show(N: in String; P: in Param);
  function Max_Size_Legend(P: Param_List) return Positive;
  function Max_Size_Chain(P: Param_List) return Positive;

end Params;
